# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import re
from enum import Enum
from pandas import to_datetime as pddt
from Octeres.SL.location_to_mask_graph import XLocation_to_Mask_None
from Octeres.SL.abstract_location_to_mask import ABC_Location_to_Mask_Graph

_, MACHINE_NAME = __name__.rsplit("ltm_", 1)


class Obj_Type(Enum):
    ROOT = "root"
    HOSTNAME = "hostname"
    IP = "ip"
    NID = "nid"
    CNAME = "cname"
    CABINET = "cabinet"
    CHASSIS = "chassis"
    SLOT = "slot"
    BLADE = "blade"
    ROW = "row"
    RACK = "rack"
    NODE = "node"
    CGROUP = "cgroup"


class Location_to_Mask_V0(ABC_Location_to_Mask_Graph):
    """This machine is what we thought polaris would look like."""

    machine_name = MACHINE_NAME
    bit_total = 112
    name_prefix = "x3"
    names = [
        (
            (Obj_Type.ROW.value, (0, 1, 1), "x3{row}"),  # row
            (Obj_Type.RACK.value, (0, 1, 1), "x3{row}{rack:0>2}"),  # rack
            (Obj_Type.CHASSIS.value, (0, 7, 1), "x3{row}{rack:0>2}c{chassis}"),
            (Obj_Type.SLOT.value, (0, 1, 1), "x3{row}{rack:0>2}c{chassis}s{slot}"),
            (
                Obj_Type.BLADE.value,
                (0, 2, 1),
                "x3{row}{rack:0>2}c{chassis}s{slot}b{blade}",
            ),  # controller, blade
            (
                Obj_Type.NODE.value,
                (0, 1, 1),
                "x3{row}{rack:0>2}c{chassis}s{slot}b{blade}n{node}",
            ),
            # (Obj_Type.CGROUP.value, (0, 8, 1), 'x3{row}{rack:0>2}c{chassis}s{slot}b{blade}n{node}[{cgroup}]'),  # from cgroups in pbs
        ),
        (
            (Obj_Type.ROW.value, (1, 2, 1), "x3{row}"),  # row
            (Obj_Type.RACK.value, (70, 73, 1), "x3{row}{rack:0>2}"),  # rack
            (Obj_Type.CHASSIS.value, (0, 7, 1), "x3{row}{rack:0>2}c{chassis}"),
            (Obj_Type.SLOT.value, (0, 1, 1), "x3{row}{rack:0>2}c{chassis}s{slot}"),
            (
                Obj_Type.BLADE.value,
                (0, 2, 1),
                "x3{row}{rack:0>2}c{chassis}s{slot}b{blade}",
            ),  # controller, blade
            (
                Obj_Type.NODE.value,
                (0, 1, 1),
                "x3{row}{rack:0>2}c{chassis}s{slot}b{blade}n{node}",
            ),
            # ('cgroup', (0, 8, 1), 'x3{row}{rack:0>2}c{chassis}s{slot}b{blade}n{node}[{cgroup}]'),  # from cgroups in pbs
        ),
        (
            (Obj_Type.ROW.value, (2, 3, 1), "x3{row}"),  # row
            (Obj_Type.RACK.value, (0, 4, 1), "x3{row}{rack:0>2}"),  # rack
            (Obj_Type.CHASSIS.value, (0, 7, 1), "x3{row}{rack:0>2}c{chassis}"),
            (Obj_Type.SLOT.value, (0, 1, 1), "x3{row}{rack:0>2}c{chassis}s{slot}"),
            (
                Obj_Type.BLADE.value,
                (0, 2, 1),
                "x3{row}{rack:0>2}c{chassis}s{slot}b{blade}",
            ),  # controller, blade
            (
                Obj_Type.NODE.value,
                (0, 1, 1),
                "x3{row}{rack:0>2}c{chassis}s{slot}b{blade}n{node}",
            ),
            # ('cgroup', (0, 8, 1), 'x3{row}{rack:0>2}c{chassis}s{slot}b{blade}n{node}[{cgroup}]'),  # from cgroups in pbs
        ),
    ]
    re_row = re.compile("^x3[0-9]$")
    re_rack = re.compile("^x3(?P<row>[0-9])(?P<rack>[0-9]{2})$")
    re_chassis = re.compile("^x3[0-9][0-9]{2}c[0-9]$")
    re_slot = re.compile("^x3[0-9][0-9]{2}c[0-9]s[0-9]$")
    re_blade = re.compile("^x3[0-9][0-9]{2}c[0-9]s[0-9]b[0-9]$")
    re_node = re.compile("^x3[0-9][0-9]{2}c[0-9]s[0-9]b[0-9]n[0-9]$")
    re_cgroup = re.compile(r"^x3[0-9][0-9]{2}c[0-9]s[0-9]b[0-9]n[0-9]\[[0-9]\]$")
    obj_types = [
        Obj_Type.ROOT.value,
        Obj_Type.ROW.value,
        Obj_Type.RACK.value,
        Obj_Type.CHASSIS.value,
        Obj_Type.SLOT.value,
        Obj_Type.BLADE.value,
        Obj_Type.NODE.value,
    ]

    def get_parents(self, machine_name: str, node_name: str) -> list:
        if machine_name == node_name:
            parents = []
        else:
            obj_type = self.get_obj_type(machine_name, node_name)
            if obj_type == Obj_Type.ROW.value:
                parents = [machine_name]
            elif obj_type == Obj_Type.RACK.value:
                match = re.match(self.re_rack, node_name)
                match_dct = match.groupdict()
                row = match_dct["row"]
                rack = match_dct["rack"]
                parents = ["%s%s" % (self.name_prefix, row)]
            elif obj_type == Obj_Type.CHASSIS.value:
                head, tail = node_name.rsplit("c", 1)
                parents = [head]
            elif obj_type == Obj_Type.SLOT.value:
                head, tail = node_name.rsplit("s", 1)
                parents = [head]
            elif obj_type == Obj_Type.BLADE.value:
                head, tail = node_name.rsplit("b", 1)
                parents = [head]
            elif obj_type == Obj_Type.NODE.value:
                head, tail = node_name.rsplit("n", 1)
                parents = [head]
            elif obj_type == Obj_Type.CGROUP.value:
                head, tail = node_name.rsplit("[", 1)
                parents = [head]
            else:
                parents = [machine_name]
        return parents

    def get_obj_type(self, machine_name: str, node_name: str) -> str:
        if machine_name == node_name:
            obj_type = Obj_Type.ROOT.value
        else:
            if re.match(self.re_row, node_name) is not None:
                obj_type = Obj_Type.ROW.value
            elif re.match(self.re_rack, node_name) is not None:
                obj_type = Obj_Type.RACK.value
            elif re.match(self.re_chassis, node_name) is not None:
                obj_type = Obj_Type.CHASSIS.value
            elif re.match(self.re_slot, node_name) is not None:
                obj_type = Obj_Type.SLOT.value
            elif re.match(self.re_blade, node_name) is not None:
                obj_type = Obj_Type.BLADE.value
            elif re.match(self.re_node, node_name) is not None:
                obj_type = Obj_Type.NODE.value
            elif re.match(self.re_cgroup, node_name) is not None:
                obj_type = Obj_Type.CGROUP.value
            else:
                raise Exception(f"Unknown obj_type, {node_name=}, {machine_name=}")
        return obj_type


ltm_lst = [
    {
        "ts": pddt("1970-01-01"),
        "te": pddt("2021-08-25"),
        "ltm_class": XLocation_to_Mask_None,
        "name": MACHINE_NAME,
    },
    {
        "ts": pddt("2021-08-25"),
        "te": pddt("2038-01-01"),
        "ltm_class": Location_to_Mask_V0,
        "pk": 39766,
        "name": MACHINE_NAME,
    },
]
