# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

from typing import Tuple, Union

import cython

from Octeres.bitmask.bitmask_globals import __zero_chunk__
from Octeres.util import superprint


def process_np_chunk(chunk: bytes, chunk_size: cython.int, chunk_step: cython.int) -> Tuple[int, int]:
    """
    Do not call this function directly unless you understand the underlying bytes structure of the chunk
    you are passing in.

    Given a chunk of bytes and a chunk_size, loop through the chunk counting bits that
    are set, length, and create a new value bitmask unsigned 64bit long.
    The bit_count is the number of bits != 0:
    The length is the length that the chunk should be.  Chunks come in chunk_size, but
    the length of the chunk could be smaller.  If there were a bitmask of 68, the first
    chunk would be 64, the second would be 4.  We must keep shifting until 64 to keep the
    set bits all the way to the left.
    """
    # valid_chunk_types = [bytes, bytearray, memoryview]
    # assert type(chunk) in valid_chunk_types, f"{type(chunk)=} not in {valid_chunk_types}"
    # superprint(f"{chunk=} {chunk_size=} {chunk_step=}")
    bit_count: cython.int = 0
    value: cython.ulonglong = 0
    start_shifting: bool = False
    idx: cython.int
    for idx in range(0, chunk_size):
        chunk_idx = idx * chunk_step
        if start_shifting:
            value = value << 1
        try:
            val = chunk[chunk_idx]
        except IndexError:
            value = value << (chunk_size - idx - 1)
            # superprint(f"chunk indexerror at {chunk_idx}, shifting {bin(value)=} << {shift_amount=}")
            break
        else:
            if val != 0:
                # only start shifting when we hit the first bit.
                start_shifting = True
                bit_count += 1
                value += 1
        # superprint(f"{idx=} {val=} {bin(value)=} {bit_count=} {start_shifting=}")
    # superprint(f"bin:{bin(value)} value:{value} bit_count:{bit_count}, chunk_length:{chunk_length}", jump_back_count=3)
    return value, bit_count


# this func needs to be outside bitmask.py as multiprocessing will need to import it.
def update_internal_mask_and_bit_count_inner_int16(binchunk: bytes, chunk_size: int, chunk_step: int):
    if binchunk == __zero_chunk__:
        _value, _bit_count = 0, 0
        # superprint(f"zero {_value=}, {_bit_count=}, {_length=}", parallel=True)
    else:
        _value, _bit_count = process_np_chunk(binchunk, chunk_size, chunk_step)
        # superprint(f"calc {_value=}, {_bit_count=}, {_length=}", parallel=True)
    return _value, _bit_count


def update_internal_mask_and_bit_count_divide_memory(memory, big_idx, chunk_size, chunk_step):
    """this requires SharedMemory to be sent over through multiprocessing"""
    # superprint('update_internal_mask_and_bit_count_divide_memory')
    binfullstep = memory.buf
    """
    index 0, offset  0 extent 64
    index 1, offset 64, extent 128
    we need
    index 1, offset 64, extent 65
    """
    offset = big_idx * chunk_size * chunk_step
    extent = (big_idx * chunk_size * chunk_step) + (chunk_size * chunk_step)
    binchunk = binfullstep[offset:extent]
    _value, _bit_count = update_internal_mask_and_bit_count_inner_int16(binchunk, chunk_size, 2)
    # superprint(f"{_bit_count=} {_length=} {_value=}")
    return _value, _bit_count


def set_all_ones(memory, big_idx, chunk_size):
    chunk_step = 2
    binfullstep = memory.buf[::chunk_step]
    # if we have 200 bits set, we are on idx 3 we only have 56 bits left, not 64
    extent = len(binfullstep) - (big_idx * chunk_size)
    if extent > chunk_size:
        extent = chunk_size
    binfullstep[big_idx * chunk_size : (big_idx * chunk_size) + chunk_size] = extent * b"\x01"


def set_all_zeros(memory, big_idx, chunk_size):
    chunk_step = 2
    binfullstep = memory.buf[::chunk_step]
    # if we have 200 bits set, we are on idx 3 we only have 56 bits left, not 64
    extent = len(binfullstep) - (big_idx * chunk_size)
    if extent > chunk_size:
        extent = chunk_size
    binfullstep[big_idx * chunk_size : (big_idx * chunk_size) + chunk_size] = extent * b"\x00"


def sleeping(idx):
    superprint(f"{idx}:sleeping")
    import time, random

    time.sleep(random.random())
    superprint(f"{idx}:done")


def get_bit_count_from_hexmask(hexmask: Union[str, bytes]) -> int:
    """calculate the number of bits in a hexmask string"""
    if type(hexmask) == bytes:
        hexmask = hexmask.decode("utf8")
    bit_count = sum([bin(int(c, 16)).count("1") for c in hexmask])
    return bit_count
