# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import dataclasses
import datetime
import math
import os
import queue
import random
import textwrap
import time
import uuid
from collections import OrderedDict, deque
from colorsys import hsv_to_rgb
from functools import reduce, wraps
from itertools import product
from typing import Tuple, List, Dict, Union, Callable, Optional

# from collections.abc import Callable

import numpy as np
import pandas as pd
from Octeres.random_generation import FixedMatrixDistribution
from Octeres.bitmask import EnhancedBitmask, EnhancedBitmaskLike
from Octeres.util import superprint

np.set_printoptions(threshold=np.inf, linewidth=np.inf)


def get_uuid() -> uuid.UUID:
    return uuid.uuid1()


def get_origin_uuid() -> uuid.UUID:
    return uuid.UUID(hex="ffffffffffffffffffffffffffffffff")


def get_null_uuid() -> uuid.UUID:
    return uuid.UUID(hex="00000000000000000000000000000000")


# FIXME: evaluate dropping into a class to have matching methods:
#   "who is my parent" functions
#   "identify which level for a string" function
# TODO:
#   add the ability to provide extra data to the formatting, such as machine_name
def _generate_names(
    names: Tuple[Tuple[str, Union[int, Tuple[int, int, int]], Union[str, Callable]], ...],
    reverse=True,
    slice_range=False,
    generate_parents=True,
    preserve_lookup=False,
    skip_nodes=(),
    extra_dct: dict = None,
) -> Tuple[List[str], Dict[str, List[str]]]:
    r"""generate a list of names based upon a template formatted input.

    names = (('a', 2, 'a{a}',           {}),
             ('b', 3, 'a{a}-b{b}',      {}),
             ('c', 5, 'a{a}-b{b}-c{c}', {}))
               ^   ^   ^                ^ TODO extra data to add to format for this level.
               |   |   | format template
               |   \ count of indexes in that level
               \ key for the name with value is an index

    or with slice.  All must be a slice if you use slice_range=True

    names = (('a', (0, 2, 1), 'a{a}'),
             ('b', (0, 3, 1), 'a{a}-b{b}'),
             ('c', (0, 5, 1), 'a{a}-b{b}-c{c}'))

    if generate_parents is on, it will probably call template.format twice.  template can be a function.
    It currently cannot be cached because there is no key, calling template, creates the key.

    for the second entry of b, it references a, so a must be before b.
    return: a tuple of
    a deepest first ordered dictionary output with the key being the level index and value being a list of names
    a deepest first ordered dictionary that maps level index to level name.
    """
    if slice_range:
        counts = [(m, n, s) for _, (m, n, s), _ in names]
    else:
        counts = [n for _, n, _ in names]
    idx_lst_of_lst = [counts[:c] for c in list(range(len(counts), 0, -1))]
    # calculate the correct count by multiplying each list of indexes then adding up the sums of a reducing list of indexes.
    if len(idx_lst_of_lst) == 0:
        correct_count = 0
    else:
        if slice_range:

            def get_slice_total(s):
                if type(s) == tuple:
                    (m, n, s) = s
                    value = len(range(*(m, n, s)))
                else:
                    value = s
                return value

            def get_total(m, n):
                mvalue = get_slice_total(m)
                nvalue = get_slice_total(n)
                return mvalue * nvalue

            q = list(
                map(
                    lambda x: (get_slice_total(x[0]) if len(x) == 1 else reduce(get_total, x)),
                    idx_lst_of_lst,
                )
            )
            correct_count = reduce(lambda c, d: c + d, q)
        else:
            correct_count = reduce(
                lambda c, d: c + d,
                map(lambda x: reduce(lambda a, b: a * b, x), idx_lst_of_lst),
            )
    idx = 0
    level_idx_dct = OrderedDict()

    parents_dct: Dict[str, List] = {}
    # max_level_depth_dct = {}
    # for lidx, (level_str, slice_tuple, template) in enumerate(names):
    #     if type(slice_tuple) == tuple:
    #         offset = slice_tuple[0]
    #     else:
    #         offset = 0
    #     max_level_depth_dct[level_str] = offset + len(range(*slice_tuple))
    # loop over each level
    preserve_lookup = {}
    for lidx, level in enumerate(names):
        level_str, _, template = level
        if slice_range:
            ranges = [list(range(*(m, n, s))) for _, (m, n, s), _ in names[: lidx + 1]]
        else:
            ranges = [list(range(m)) for _, m, _ in names[: lidx + 1]]
        # loop over each range in the level
        level_name_lst = []
        for ridx, row in enumerate(product(*ranges)):  # the cartesian product of the ranges
            name_dct = {"idx": ridx, "lvl": lidx}
            if extra_dct:
                name_dct.update(extra_dct)
            # loop over each column in the row and create the dictionary for the template.
            for cidx, value in enumerate(row):
                key_name = names[cidx][0]
                name_dct[key_name] = value
                name_dct[f"{key_name}__str"] = key_name
                # name_dct[f"{key_name}__max"] = max_level_depth_dct[key_name]
            if type(template) == str:
                name = template.format(**name_dct, fill="0")
            else:
                name = template(name_dct)
            if name not in skip_nodes:
                level_name_lst.append(name)
                preserve_lookup[name] = name_dct
            if generate_parents and lidx > 0:
                template_parent = names[lidx - 1][2]
                if type(template_parent) == str:
                    parents = [template_parent.format(**name_dct, fill="0")]
                else:
                    parents = template_parent(name_dct, generate_parent=True)
                if name not in skip_nodes:
                    if type(parents) != list:
                        parents = [parents]
                    parents_dct[name] = parents
            idx += 1
        level_idx_dct[lidx] = level_name_lst
    assert idx == correct_count
    # level_idx_ordereddct: OrderedDict[int, List[str]] = OrderedDict()
    nodes = []
    # the dictionary must be walked through backwards from the bottom back up to the top.
    if reverse:
        for lidx in reversed(level_idx_dct.keys()):
            nodes.extend(level_idx_dct[lidx])
            # level_idx_ordereddct[lidx] = level_idx_dct[lidx]
    else:
        for lidx in sorted(level_idx_dct.keys()):
            nodes.extend(level_idx_dct[lidx])
            # level_idx_ordereddct[lidx] = level_idx_dct[lidx]
    # return level_idx_ordereddct
    return nodes, parents_dct


def generate_names(
    names: Union[
        Tuple,
        List[Tuple[Tuple[str, Union[int, Tuple[int, int, int]], Union[str, Callable]], ...]],
    ],
    reverse: bool = True,
    slice_range: bool = False,
    generate_parents: bool = True,
    skip_nodes=(),
    extra_dct={},
) -> Tuple[List[str], Dict[str, List[str]]]:
    if not names:
        nodes = []
        parents_dct = {}
    else:
        parents_dct = {}
        if type(names) == list and names:
            nodes: List[str] = []
            for name_tup in names:
                names_tmp, parents_dct_tmp = _generate_names(
                    name_tup,
                    reverse=reverse,
                    slice_range=slice_range,
                    generate_parents=generate_parents,
                    skip_nodes=skip_nodes,
                    extra_dct=extra_dct,
                )
                nodes.extend(names_tmp)
                parents_dct.update(parents_dct_tmp)
        elif type(names) == tuple and names:
            nodes, parents_dct = _generate_names(
                names,
                reverse=reverse,
                slice_range=slice_range,
                generate_parents=generate_parents,
                skip_nodes=skip_nodes,
                extra_dct=extra_dct,
            )
        else:
            raise Exception(f"Unsupported {type(names)=}, should be a list or tuple")
    return nodes, parents_dct


def format_array(out: np.ndarray):
    box = []
    inner_box = []
    first = True
    for x_idx, x_arr in enumerate(out):
        for y_idx, y in enumerate(x_arr.tolist()):
            y_str = y
            if first:
                # y or nodes
                inner_box.append(y_str)
            else:
                inner_box[y_idx] = inner_box[y_idx] + y_str
        first = False
    box.extend(inner_box)
    x_length = len(inner_box[0])
    box.insert(0, "-" * x_length)

    def number_generator(length):
        idx = 0
        for _ in range(length):
            yield idx
            idx += 1
            if idx > 9:
                idx = 0

    g = number_generator(x_length)
    box.append("-" * x_length)
    box.append("".join([str(next(g)) for _ in range(x_length)]))
    box_str = "\n".join(box)
    box_str = textwrap.indent(box_str, "    |")
    return box_str


class EventGeneration:
    def __init__(
        self,
        size_x,
        size_y,
        dtype="U3",
        empty_value="   ",
        test_value="...",
        visualize=False,
        bitmask_class=EnhancedBitmask,
    ):
        self.size_x = size_x  # time
        self.size_y = size_y  # nodes
        self.dtype = dtype
        self.empty_value = empty_value
        self.test_value = test_value
        out = np.zeros((size_x, size_y), dtype=self.dtype)
        out[:, :] = self.empty_value
        # out[1:, 0] = "x"
        # out[0, 1:] = "y"
        self.out = out
        self.empty = out.copy()
        self.visualize = visualize
        self.bitmask_class: EnhancedBitmaskLike = bitmask_class

    def factor_00(self, area, max_x, max_y) -> Tuple[int, int]:
        x = round(1 + (area / (random.randint(3, max_y))))
        y = round(1 + (area / x))
        x = x if x < max_x else max_x
        y = y if y < max_y else max_y
        return x, y

    def factor_01(self, area, max_x, max_y) -> Tuple[int, int]:
        # the 1 + exists, because area(numerator) can be less than the denominator.
        y = round(area / (random.randint(1, max_y)))
        try:
            x = round(area / y)
        except ZeroDivisionError:
            x = area
        if y == 0:
            y = 1
        if x == 0:
            x = 1
        x = x if x < max_x else max_x
        y = y if y < max_y else max_y
        return x, y

    def factor_02(self, area, max_x, max_y) -> Tuple[int, int]:
        """given an area to fit, come up with a random box."""
        s = math.sqrt(area)
        scale_factor_a = random.randint(60, 160) / 100
        # if x is heavy, then we will lean the scale factor towards it
        x_heavy = 0 if max_x > max_y else 1
        y_heavy = 0 if max_y > max_x else 1
        x_scaled = s * abs(x_heavy - scale_factor_a)  # 1 - will lead to a smaller percentage due to the 60-100 range.
        y_scaled = s * abs(y_heavy - scale_factor_a)
        scale_factor_b = random.randint(-40, 40) / 100  # adjust by n percent
        x = x_scaled + (x_scaled * scale_factor_b)
        y = y_scaled + (y_scaled * (1 - scale_factor_b))
        x = math.ceil(x)
        y = math.ceil(y)
        x = x if x < max_x else max_x
        y = y if y < max_y else max_y
        x = 1 if x < 1 else x
        y = 1 if y < 1 else y
        return x, y

    def check_box(self, out, x, y, x_ext, y_ext, test_value=None):
        if test_value is None:
            test_value = self.test_value

        subarray = out[x : x + x_ext, y : y + y_ext]
        clear = np.all(subarray == self.empty_value)

        if self.visualize:
            if clear:
                original_subarray = subarray.copy()
                subarray[:, :] = test_value
                print(format_array(out))
                subarray[:, :] = original_subarray
            else:
                print(format_array(out))
        return clear

    def check_box_v0(self, out, x, y, x_ext, y_ext, test_value=None):
        if test_value is None:
            test_value = self.test_value
        if self.visualize:
            # os.system('clear')
            # print_array(self.empty)
            out_copy: np.ndarray = out.copy()
        clear = True
        for x_idx, x_val in enumerate(out[x : x + x_ext, y : y + y_ext]):
            if self.visualize:
                out_copy[x : x + x_ext, y : y + y_ext] = test_value
            for y_idx, y_val in enumerate(x_val):
                if y_val != self.empty_value:
                    clear = False
                    break
        if self.visualize:
            print(format_array(out_copy))
        return clear

    def assign_box(self, out, x, y, x_ext, y_ext, value):
        out[x : x + x_ext, y : y + y_ext] = value
        if self.visualize:
            print(format_array(out))

    def get_a_random_corner_00(self, out, max_x, max_y):
        """
        this gets a random corner, but then moves up and to the left a bit.
        It could walk up until it hits a will and switch to the other dimension (x vs y)
        """
        possible_x = random.randint(0, max_x)
        possible_y = random.randint(0, max_y)
        # let's check up and to the left a bit.
        corner_x = possible_x
        corner_y = possible_y
        while corner_x >= 0 and corner_y >= 0:
            if out[corner_x, corner_y] == self.empty_value:
                possible_x = corner_x
                possible_y = corner_y
                if corner_x >= 0:
                    corner_x -= 1
                if corner_y >= 0:
                    corner_y -= 1
            else:
                break
        return possible_x, possible_y

    def get_a_random_corner_01(self, out, max_x, max_y):
        """
        this gets a random corner, but then moves up and to the left a bit.
        It could walk up until it hits a will and switch to the other dimension (x vs y)
        """

        # possible_x = random.randint(0, max_x)
        # possible_y = random.randint(0, max_y)
        possible_x = int(max_x * random.random())
        possible_y = int(max_y * random.random())

        return possible_x, possible_y

    def move_around_a_bit_v0(
        self,
        out,
        possible_x,
        possible_y,
        x_ext,
        y_ext,
        x_direction,
        y_direction,
        max_x,
        max_y,
    ):
        # let's check up and to the left a bit.
        corner_x = possible_x
        corner_y = possible_y
        # FIXME: when we check the corner, we need to check based upon the direction x or y
        # first go up until you hit something, stop
        # then go back until you hit something, stop
        while corner_x >= 0 and corner_x < max_x:
            if x_direction == 1:
                value = out[corner_x + x_ext - 1, corner_y]
            else:
                value = out[corner_x, corner_y]
            if value == self.empty_value:
                possible_x = corner_x
                corner_x += x_direction
            else:
                break
        while corner_y >= 0 and corner_y < max_y:
            if y_direction == 1:
                value = out[corner_x, corner_y + y_ext - 1]
            else:
                value = out[corner_x, corner_y]
            if value == self.empty_value:
                possible_y = corner_y
                corner_y += y_direction
            else:
                break
        return possible_x, possible_y

    def move_around_a_bit(
        self,
        out,
        possible_x,
        possible_y,
        x_ext,
        y_ext,
        x_direction,
        y_direction,
        max_x,
        max_y,
    ):
        # Cut some redundant moves in move_around_a_bit
        corner_x = possible_x
        corner_y = possible_y

        if x_direction == 1:
            x_range = range(possible_x, min(possible_x + x_ext, max_x))
        else:
            x_range = range(possible_x, max(possible_x - x_ext, -1), -1)

        if y_direction == 1:
            y_range = range(possible_y, min(possible_y + y_ext, max_y))
        else:
            y_range = range(possible_y, max(possible_y - y_ext, -1), -1)

        for corner_x in x_range:
            if out[corner_x, corner_y] != self.empty_value:
                break
            possible_x = corner_x

        for corner_y in y_range:
            if out[corner_x, corner_y] != self.empty_value:
                break
            possible_y = corner_y

        return possible_x, possible_y

    def find_location_00(self, out: np.ndarray, value, x_ext, y_ext):
        # a really quick scheduling algorithm
        shape_x, shape_y = out.shape
        max_x = shape_x - x_ext
        max_y = shape_y - y_ext
        assert max_x >= 0
        assert max_y >= 0
        try_count = 0
        x = 0
        y = 0
        found = False
        # find a corner, walk y
        for possible_y in range(0, max_y):
            possible_x = 0
            clear = self.check_box(out, possible_x, possible_y, x_ext, y_ext)
            if clear:
                self.assign_box(out, possible_x, possible_y, x_ext, y_ext, value)
                x, y = possible_x, possible_y
                found = True
                break

        if not found:
            for possible_x in range(0, max_x):
                possible_y = 0
                clear = self.check_box(out, possible_x, possible_y, x_ext, y_ext)
                if clear:
                    self.assign_box(out, possible_x, possible_y, x_ext, y_ext, value)
                    x, y = possible_x, possible_y
                    found = True
                    break

        if not found:
            while try_count < 16:
                possible_x, possible_y = self.get_a_random_corner_00(out, max_x, max_y)
                clear = self.check_box(out, possible_x, possible_y, x_ext, y_ext)
                if clear:
                    self.assign_box(out, possible_x, possible_y, x_ext, y_ext, value)
                    x, y = possible_x, possible_y
                    found = True
                    break
                try_count += 1
        return found, x, y

    def find_location_01(self, out: np.ndarray, value, x_ext, y_ext):
        # a really quick scheduling algorithm
        # this makes a weird angled pattern like stalagtites
        shape_x, shape_y = out.shape
        max_x = shape_x - x_ext
        max_y = shape_y - y_ext
        superprint(f"{value=} {x_ext=} {y_ext=}")
        assert max_x >= 0
        assert max_y >= 0
        try_count = 0
        x = 0
        y = 0
        found = False
        if not found:
            while try_count < 16:
                possible_x, possible_y = self.get_a_random_corner_00(out, max_x, max_y)
                clear = self.check_box(out, possible_x, possible_y, x_ext, y_ext)
                if clear:
                    self.assign_box(out, possible_x, possible_y, x_ext, y_ext, value)
                    x, y = possible_x, possible_y
                    found = True
                    break
                try_count += 1
        return found, x, y

    def find_location_02(self, out: np.ndarray, value, x_ext: int, y_ext: int):
        # a really quick scheduling algorithm
        # this makes a weird angled pattern like stalagtites
        shape_x, shape_y = out.shape
        # random event generate is x length = 4 and y length = 3
        # x_ext = 4, y_ext = 3
        # max_x = 16 - 4 => 12
        # max_y = 7 - 3 => 4
        # i can place the top left corner up to 12x4
        #  xxxx xxxx xxxx xxxx
        # y                |
        # y                |
        # y                |
        # y                |
        # y________________|
        # y                |
        # y                |

        max_x = shape_x - x_ext
        max_y = shape_y - y_ext
        assert max_x >= 0
        assert max_y >= 0
        x_direction = -1 if random.randint(0, 1) == 0 else 1
        y_direction = -1 if random.randint(0, 1) == 0 else 1

        try_count = 0
        x = 0
        y = 0
        found = False
        if not found:
            while try_count < 16:
                possible_x, possible_y = self.get_a_random_corner_01(out, max_x, max_y)
                possible_x, possible_y = self.move_around_a_bit(
                    out,
                    possible_x,
                    possible_y,
                    x_ext,
                    y_ext,
                    x_direction,
                    y_direction,
                    max_x,
                    max_y,
                )
                clear = self.check_box(
                    out,
                    possible_x,
                    possible_y,
                    x_ext,
                    y_ext,
                    test_value=self.test_value,
                )
                if clear:
                    self.assign_box(out, possible_x, possible_y, x_ext, y_ext, value)
                    x, y = possible_x, possible_y
                    found = True
                    break
                try_count += 1
        return found, x, y

    def generate_names_2deep(self, count):
        """limited to 3844, 2 wide"""
        event_name_generator = list(range(97, 122 + 1)) + list(range(65, 90 + 1)) + list(range(48, 57 + 1))
        names = list(map(chr, event_name_generator))
        max_names = len(names) * len(names)
        if count > max_names:
            raise Exception(f"generate name count limit reached. {max_names}")
        event_names = []
        for oname in names:
            for iname in names:
                event_names.append(f"{oname}{iname}")
        return event_names[:count]

    def generate_names_3deep(self, count) -> List[str]:
        """up to 238328 deep, 3 wide"""
        event_name_generator = list(range(97, 122 + 1)) + list(range(65, 90 + 1)) + list(range(48, 57 + 1))
        names = list(map(chr, event_name_generator))
        max_names = len(names) * len(names) * len(names)
        if count > max_names:
            raise Exception(f"generate name count limit reached. {max_names}")
        event_names = []
        for aname in names:
            for bname in names:
                for cname in names:
                    event_names.append(f"{aname}{bname}{cname}")
        return event_names[:count]

    def generate_names_n_level(self, characters, width) -> List[str]:
        """generate count ** width names"""
        lstoflst = (characters for _ in range(width))
        event_names = ["".join(t) for t in product(*lstoflst)]
        return event_names

    def generate_non_overlapping_scattered_events_v0(self, event_names: List[str], subfigure_count: Optional[int] = None):
        """Generate non-overlapping events that covers most of the timeline. Each event has random multiple intervals
        The distribution of intervals is decided in FixedMatrixDistribution, which is defined in random_generation.py """
        size_x = self.size_x
        size_y = self.size_y
        big_figure, offset_x, offset_y = self.generate_random_big_figure(size_x, size_y)
        sizes_xy = self.divide_figure(
            big_figure.shape[1],
            big_figure.shape[0],
            len(event_names),
            subfigure_count=subfigure_count,
        )
        random.shuffle(sizes_xy)
        # event_queue = queue.Queue()
        # [event_queue.put(l) for l in event_names]
        num_events_to_process = min(len(sizes_xy), len(event_names)) // 15
        event_queue = queue.Queue()
        for name in event_names[:num_events_to_process]:
            event_queue.put(name)

        event_lst = []
        base_mask = self.bitmask_class.zeros(size_y)
        event_not_lst = []
        for name in event_names:
            event_queue.put(name)
        rd = FixedMatrixDistribution()
        while not event_queue.empty() and sizes_xy:
            value = event_queue.get()
            count = rd.get_random_number()
            bitmask_range = []

            while count > 0 and sizes_xy:
                x_ext, y_ext = sizes_xy.pop(0)
                found, x, y = self.find_location_02(big_figure, value, x_ext, y_ext)
                if found:
                    big_figure[y : y + y_ext, x : x + x_ext] = 0
                    bitmask_range.extend(list(range(y + offset_y, y + y_ext + offset_y)))
                    count -= 1
                if count == 0:
                    event_bitmask = self.bitmask_class.zeros_and_set_bits(base_mask.length, bitmask_range)
                    event_dct = {
                        "name": value,
                        "x": x + offset_x,
                        "y": y + offset_y,
                        "ts": x + offset_x,
                        "te": x + offset_x + x_ext,
                        "bitmask": event_bitmask,
                    }
                    event_lst.append(event_dct)

        events = GeneratedEvents(big_figure, event_lst, event_not_lst)
        return events

    def generate_non_overlapping_box_events_v0(self, event_names: List[str]):
        """Generate non-overlapping events that covers most of the timeline. Each event have only one interval."""
        size_x = self.size_x
        size_y = self.size_y
        out = self.out.copy()
        """
        x is time, the outer loop
        y is the mask, or nodes
        """
        len_events = len(event_names)
        total_area = size_x * size_y
        event_queue = queue.Queue()
        [event_queue.put(l) for l in event_names]

        big_factor_area = 0.60
        big_factor_count = 0.20
        big_area = round(big_factor_area * total_area)
        big_count = round(len_events * big_factor_count)
        big_avg = round(big_area / big_count)  # this could be done using field_y and field_x and a generator for big_sizes
        big_sizes = np.random.randint(round(big_avg / 2), round(big_avg * 2), size=big_count).tolist()
        # big_sizes = [random.randint(round(big_avg / 2), round(big_avg * 2)) for a in range(big_count)]
        big_sizes_xy = [self.factor_02(size, size_x, size_y) for size in big_sizes]
        big_sizes_xy = sorted(big_sizes_xy, key=lambda t: t[0], reverse=True)
        small_count = (len_events - big_count) * 2  # lots of little events.
        small_area = (total_area - big_area) + (total_area - big_area) * 0.5
        small_avg = round(small_area / small_count)
        if small_avg < 1:
            small_avg = 1
        small_sizes = [random.randint(1, small_avg) for a in range(small_count)]
        small_sizes_xy = [self.factor_02(size, size_x, size_y) for size in small_sizes]
        small_sizes_xy = sorted(small_sizes_xy, key=lambda t: t[1], reverse=True)

        sizes_xy = big_sizes_xy + small_sizes_xy
        event_lst = []
        base_mask = self.bitmask_class.zeros(size_y)

        event_idx = 0
        event_not_lst = []
        while not event_queue.empty() and (value := event_queue.get()):
            x_ext, y_ext = sizes_xy.pop(0)
            found, x, y = self.find_location_02(out, value, x_ext, y_ext)
            if found:
                # event_bitmask = base_mask.copy()
                # event_bitmask[y:y + y_ext] = 1
                # other_bitmask = EnhancedBitmask.zeros(base_mask.length)
                # other_bitmask[y:y + y_ext] = 1
                event_bitmask = self.bitmask_class.zeros_and_set_bits(base_mask.length, [_ for _ in range(y, y + y_ext)])
                event_dct = {
                    "name": value,
                    "x": x,
                    "y": y,
                    "ts": x,
                    "te": x + x_ext,
                    "bitmask": event_bitmask,
                }
                event_lst.append(event_dct)
            else:
                pass
                event_not_lst.append(value)
            event_idx += 1
        events = GeneratedEvents(out, event_lst, event_not_lst)
        return events

    def generate_non_overlapping_box_events_v1(self, event_names: List[str], subfigure_count: Optional[int] = None):
        """Generate non-overlapping box events, using rectangle division algorithm to distribute intervals"""
        size_x = self.size_x
        size_y = self.size_y
        big_figure, offset_x, offset_y = self.generate_random_big_figure(size_x, size_y)
        sizes_xy = self.divide_figure(
            big_figure.shape[1],
            big_figure.shape[0],
            len(event_names),
            subfigure_count=subfigure_count,
        )
        random.shuffle(sizes_xy)
        # event_queue = queue.Queue()
        # [event_queue.put(l) for l in event_names]
        num_events_to_process = min(len(sizes_xy), len(event_names))
        event_queue = queue.Queue()
        for name in event_names[:num_events_to_process]:
            event_queue.put(name)
        event_lst = []
        base_mask = self.bitmask_class.zeros(size_y)
        event_not_lst = []
        idx = 0
        while not event_queue.empty():
            value = event_queue.get()
            x_ext, y_ext = sizes_xy.pop(0)
            found, x, y = self.find_location_02(big_figure, value, x_ext, y_ext)
            if found:
                big_figure[y : y + y_ext, x : x + x_ext] = 0
                event_bitmask = self.bitmask_class.zeros_and_set_bits(
                    base_mask.length, list(range(y + offset_y, y + y_ext + offset_y))
                )
                event_dct = {
                    "name": value,
                    "x": x + offset_x,
                    "y": y + offset_y,
                    "ts": x + offset_x,
                    "te": x + offset_x + x_ext,
                    "bitmask": event_bitmask,
                }
                event_lst.append(event_dct)
            else:
                event_not_lst.append(value)
            idx += 1

        events = GeneratedEvents(big_figure, event_lst, event_not_lst)
        return events

    def generate_random_big_figure(self, max_x: int, max_y: int) -> Tuple[np.ndarray, int, int]:
        """ The rel_size_factor is used to control the width and height of the figure, as well as the offset of them.
        Suppose the factor is 0.9, the width and height are 90%, and the size of the big_figure is 81%.
        This means the consumed_unit_seconds/possible_unit_seconds would be no larger than 0.81."""
        rel_size_factor = random.uniform(1.0, 1.0)
        big_x = int(max_x * rel_size_factor)
        big_y = int(max_y * rel_size_factor)
        big_figure = np.full((big_x, big_y), self.empty_value, dtype=self.dtype)
        offset_x = random.randint(0, max_x - big_x)
        offset_y = random.randint(0, max_y - big_y)
        return big_figure, offset_x, offset_y

    def divide_figure(
        self,
        size_x: int,
        size_y: int,
        len_events: int,
        subfigure_count: Optional[int] = None,
    ):
        """Use the recursive division to divide figure and return a list with the width and height of each rectangle"""
        if subfigure_count is None:
            subfigure_count = len_events
        rects = []
        self.recursive_divide(0, 0, size_x, size_y, rects, subfigure_count)
        return [(rect[2], rect[3]) for rect in rects]

    def recursive_divide(self, x, y, width, height, rects, target_rects, min_size=20):
        """Depending on the width and height, divide sub-rectangles recursively at the given direction"""
        if len(rects) >= target_rects or width <= min_size or height <= min_size:
            rects.append((x, y, width, height))
            return

        if width > height:
            split_vertical = True
        elif height > width:
            split_vertical = False
        else:
            split_vertical = random.choice([True, False])

        if split_vertical:
            split_at = random.randint(int(width * 0.3), int(width * 0.7))
            self.recursive_divide(x, y, split_at, height, rects, target_rects, min_size)
            self.recursive_divide(x + split_at, y, width - split_at, height, rects, target_rects, min_size)
        else:
            split_at = random.randint(int(height * 0.3), int(height * 0.7))
            self.recursive_divide(x, y, width, split_at, rects, target_rects, min_size)
            self.recursive_divide(x, y + split_at, width, height - split_at, rects, target_rects, min_size)

    def iterative_divide(x, y, width, height, rects, target_rects, min_size=20):
        """Using stack instead of recursive divide for dividing"""
        stack = deque([(x, y, width, height)])

        while stack and len(rects) < target_rects:
            x, y, width, height = stack.pop()

            if width <= min_size or height <= min_size or len(rects) >= target_rects:
                rects.append((x, y, width, height))
                continue

            if width > height:
                split_vertical = True
            elif height > width:
                split_vertical = False
            else:
                split_vertical = random.choice([True, False])

            if split_vertical:
                split_at = random.randint(int(width * 0.3), int(width * 0.7))
                stack.append((x + split_at, y, width - split_at, height))
                stack.append((x, y, split_at, height))
            else:
                split_at = random.randint(int(height * 0.3), int(height * 0.7))
                stack.append((x, y + split_at, width, height - split_at))
                stack.append((x, y, width, split_at))


@dataclasses.dataclass
class GeneratedEvents:
    out: np.ndarray
    event_lst: list
    event_not_lst: list


class UnderflowError(ArithmeticError):
    pass


def int_min_max(int_min, int_max):
    def outer(func):
        @wraps(func)
        def inner(i, s):
            if int_min <= i < int_max:
                result = func(i, s)
            else:
                if i >= int_max:
                    raise OverflowError(f"{i} >= {int_max}")
                else:
                    raise UnderflowError(f"{i} < {int_min}")
            return result

        return inner

    return outer


INT_BITS = 24
INT_MAX = 2**INT_BITS


@int_min_max(0, INT_MAX)
def rotate_int_left(i, s):
    return (i << s) | (i >> (INT_BITS - s))


@int_min_max(0, INT_MAX)
def rotate_int_right(i, s):
    return (i >> s) | (i << (INT_BITS - s)) & (INT_MAX - 1)


def color_from_int(i):
    shift = int(bin(i)[-4:], 2)
    i = rotate_int_right(i, shift)
    return f"#{hex(i)[2:]:0>6}"


def color_from_int_with_random(i):
    shift = int(bin(i)[-4:], 2)
    i = rotate_int_right(i, shift)
    r = i >> 16
    g = ((i >> 8 << 8) - (i >> 16 << 16)) >> 8
    b = (i) - (i >> 16 << 16) - ((i >> 8 << 8) - (i >> 16 << 16))
    rgb_random = []
    for ri in [r, g, b]:
        r = random.randint(0, 32) * (-1 if random.randint(0, 1) == 0 else 1)
        ri = r + ri
        if ri < 0:
            ri = 0
        if ri > 255:
            ri = 255
        rgb_random.append(ri)
    r, g, b = rgb_random
    return f"#{hex(r)[2:]:0>2}{hex(g)[2:]:0>2}{hex(b)[2:]:0>2}"


class GenerateRandom:
    def __init__(self, columns: dict):
        self.rng = np.random.default_rng()
        self.columns = columns
        self.alphabet = list("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
        self.np_info_lookup = dict(
            f32=dict(min=np.finfo(np.single).min, max=np.finfo(np.single).max),
            f64=dict(min=np.finfo(np.double).min, max=1.7976931348623157e291),  # why?
        )
        self.get_array_size = lambda: 4

    def generate_random_column(self, definition):
        if isinstance(definition, Callable):
            value = definition()
        elif (
            type(definition) == dict
            and "metadata" in definition
            and "name" in definition
            and "nullable" in definition
            and "type" in definition
        ):
            # this most likely came from a spark schema.jsonValue()
            value_type = definition["type"]
            if value_type == "binary":
                value = np.random.choice(self.alphabet, size=4)
                value = "".join(value.tolist())
                value = value.encode("utf-8")
            elif value_type == "string":
                value = np.random.choice(self.alphabet, size=4)
                value = "".join(value.tolist())
            elif value_type == "integer":
                value = np.random.randint(-2147483648, 2147483647)
            elif value_type == "long":
                value = np.random.randint(-9223372036854775808, 9223372036854775807, dtype=np.int64)
            elif value_type == "timestamp":
                value = datetime.datetime.utcnow()
            elif value_type == "double":
                np_info = self.np_info_lookup["f64"]
                value = np.random.uniform(np_info["min"], np_info["max"])
            elif value_type == "single":
                np_info = self.np_info_lookup["f32"]
                value = np.random.uniform(np_info["min"], np_info["max"])
            elif type(value_type) == dict and "type" in value_type and "elementType" in value_type:
                sub_value_type = value_type["type"]
                et = value_type["elementType"]
                if sub_value_type == "array" and et["type"] in ["struct", "array"]:
                    lst = []
                    array_size = et["size"] if "size" in et else self.get_array_size()
                    for i in range(array_size):
                        if et["type"] == "struct":
                            sub_dct = {}
                            for field in et["fields"]:
                                sub_value = self.generate_random_column(field)
                                if sub_value is not None:
                                    sub_dct[field["name"]] = self.generate_random_column(field)
                            if sub_dct:
                                lst.append(sub_dct)
                        elif et["type"] == "array":
                            sub_lst = []
                            for field in et["fields"]:
                                sub_value = self.generate_random_column(field)
                                if sub_value is not None:
                                    sub_lst.append(sub_value)
                            if sub_lst:
                                lst.append(sub_lst)
                    if lst:
                        value = lst
                    else:
                        value = None
                else:
                    raise Exception("?")
            elif type(value_type) == dict and "type" in value_type and value_type["type"] == "struct":
                dct = {}
                for field in value_type["fields"]:
                    dct[field["name"]] = self.generate_random_column(field)
                value = dct
            # elif value_type == 'struct':
            #     pass
            # elif value_type == 'array':
            #     pass
            else:
                raise Exception(f"unsupported random type {value_type}")
        else:
            assert type(definition) == dict
            assert "a" in definition
            arraylike = definition["a"]
            size = definition.get("size", 1)
            if type(size) == tuple:
                idx, j = size
                size = random.randint(idx, j)
            probability2 = definition.get("p2", None)
            if probability2:
                p2_lst = []
                for cnt, ptotal in probability2:
                    for _ in range(cnt):
                        p2_lst.append(ptotal / cnt)
                assert sum(p2_lst) - 1 < 0.0000000001, sum(p2_lst)
                probability = p2_lst
            else:
                probability = definition.get("p", None)
            pfunc = definition.get("pfunc", lambda _: _)
            value = pfunc(self.rng.choice(arraylike, size=size, p=probability))
        return value

    def generate_random_series(self, return_dict=False) -> Union[pd.Series, dict]:
        dct = dict()
        for key, definition in self.columns.items():
            value = self.generate_random_column(definition)
            dct[key] = value
        if return_dict is False:
            result = pd.Series(dct)
        else:
            result = dct
        return result


def generate_random_dataframe(columns: dict, row_count: int):
    """
    columns: dict
        key: column name
        value:
            function to be called and return data for that column
            or
            dict with  (numpy.random._generator.Generator.choice)
                a: arraylike for direct use in choice
                size:
                    if int: for direct use in choice
                    if tuple2: passed to random.randint then to size.
                p: for direct use in choice if p2 not set.
                p2: probability tuple[tuple2] where each tuple is count and ptotal for that count.
                pfunc: function to be executed after choice on the data, useful for joining lists
    row_count: int
        number of rows to put in the dataframe.
    """
    lst = []
    gr = GenerateRandom(columns)
    for i in range(row_count):
        ser = gr.generate_random_series()
        ser["idx"] = i
        lst.append(ser)
    df = pd.DataFrame(lst)
    df.set_index("idx", inplace=True)
    return df


class RandomTimeRange:
    def __init__(
        self,
        range_start: datetime.datetime,
        range_end: datetime.datetime,
        base_unit: datetime.timedelta,
        count: int,
    ):
        self.range_start = range_start
        self.range_end = range_end
        self.base_unit = base_unit
        self.count = count  # only used by get_times_ordered

    def get_times_ordered(self):
        """produce a list of times from range_start to range_end"""
        delta = (self.range_end - self.range_start).total_seconds() / self.base_unit.total_seconds()
        numbers = sorted([(random.random() * delta) for _ in range(self.count)])
        ts_gen = (self.range_start + (n * self.base_unit) for n in numbers)
        return ts_gen

    def generate_times_range(self):
        return next(self.get_times_ordered())

    def _generator_times(self):
        curr_dt = self.range_start
        while True:
            delta = random.random() * self.base_unit.total_seconds()
            curr_dt = curr_dt + datetime.timedelta(seconds=delta)
            yield curr_dt

    def generator_times(self):
        return next(self._generator_times())


class RandomIncreasingInt:
    def __init__(self, int_min, int_max, int_min_step, int_max_step):
        self.int_min = int_min
        self.int_max = int_max
        self.int_min_step = int_min_step
        self.int_max_step = int_max_step
        self.curr = self.int_min

    def get_int(self):
        self.curr = self.curr + random.randint(self.int_min_step, self.int_max_step)
        return self.curr

    def get_int_lst(self, count=2):
        lst = []
        first = True
        for i in range(count):
            n = self.get_int()
            if first:
                # return None if the first item in the list is > max.
                if n > self.int_max:
                    lst = None
                    break
                first = False
            lst.append(n if n <= self.int_max else self.int_max)
        return lst

    def reset(self):
        self.curr = self.int_min


class NameGeneratorIDX:
    def __init__(self, count, initial_idx=0):
        self.initial_idx = initial_idx
        self.idx = self.initial_idx
        self.count = count

    def reset(self):
        self.idx = self.initial_idx

    def generate_idx(self, name_dct, generate_parent=False):
        if self.idx > self.count:
            raise Exception(f"NameGenerator overflowed count of {self.count}")
        name = f"{self.idx}"
        self.idx += 1
        return name


def generate_colors(count):
    """grab a greenish color"""
    colors = []
    for index in range(count):
        # Tweak these values to change colours/variance
        hue = uniform(0.0861, 0.6222)  # Select random green'ish hue from hue wheel
        sat = uniform(0.2, 1)
        val = uniform(0.3, 1)

        red, green, blue = hsv_to_rgb(hue, sat, val)

        # Convert to 0-1 range for HTML output
        red, green, blue = [x * 255 for x in (red, green, blue)]
        red = "%x" % int(red)
        green = "%x" % int(green)
        blue = "%x" % int(blue)
        if len(red) == 1:
            red = "0%s" % red
        if len(green) == 1:
            green = "0%s" % green
        if len(blue) == 1:
            blue = "0%s" % blue
        colors.append("#%s%s%s" % (red, green, blue))
    return colors
