# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""this will contain the database helper libraries such as
chunking inserts and such"""
import csv
import gzip
import copy
import pyodbc
import datetime
from sqlalchemy.orm.collections import InstrumentedList


def get_rows(cursor):
    """Returns all rows from a cursor into (columns, lstofdct)"""
    chunk = cursor.fetchall()
    columns = [col[0] for col in cursor.description]
    return [dict(zip(columns, row)) for row in chunk]


def get_row_with_limit(cursor, limit, columns=None):
    """Returns all rows from a cursor into (columns, lstofdct)"""
    chunk = cursor.fetchmany(limit)
    if not chunk:
        result = None
    else:
        if not columns:
            desc = cursor.description
            columns = [col[0] for col in desc]
        result = [dict(zip(columns, row)) for row in chunk]
    return result


def format_desc(desc):
    """given a pyodbc desc, return a dictionary with each name to the values."""
    lookup = {}
    for name, type_code, display_size, internal_size, precision, scale, null_ok in desc:
        if type_code == str:
            type_code = "str"
        elif type_code == datetime.datetime:
            type_code = "datetime"
        elif type_code == int:
            type_code = "int"
        elif type_code == float:
            type_code = "float"
        else:
            type_code = "str"
        lookup[name] = dict(
            name=name,
            type_code=type_code,
            display_size=display_size,
            internal_size=internal_size,
            precision=precision,
            scale=scale,
            null_ok=null_ok,
        )
    return lookup


class RowIter:
    def __init__(self, cursor, limit=10000):
        self.cursor = cursor
        self.desc = cursor.description
        self.columns = [col[0] for col in self.desc]
        self.limit = limit
        self.chunk_offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        lstofdct = get_row_with_limit(self.cursor, self.limit)
        curr_chunk_offset = self.chunk_offset
        self.chunk_offset = self.chunk_offset + self.limit
        if lstofdct is None:
            raise StopIteration
        return curr_chunk_offset, self.columns, lstofdct


def dump_table_to_csv(conn, sql, limit, output_path):
    initial = True
    row_count = 0
    if type(sql) == str:
        sql_lst = [sql]
    else:
        sql_lst = sql
    desc = None
    for sql in sql_lst:
        cursor = conn.cursor()
        cursor.execute(sql)
        for tup in RowIter(cursor, limit=limit):
            offset, columns, lstofdct = tup
            row_count += len(lstofdct)
            if initial:
                # this will overwrite any file.
                with gzip.open(output_path, "wt") as f:
                    dict_writer = csv.DictWriter(f, columns, quotechar="|", quoting=csv.QUOTE_ALL)
                    dict_writer.writeheader()
                    dict_writer.writerows(lstofdct)
                initial = False
            else:
                with gzip.open(output_path, "at") as f:
                    dict_writer = csv.DictWriter(f, columns, quotechar="|", quoting=csv.QUOTE_ALL)
                    dict_writer.writerows(lstofdct)
        desc = cursor.description
        cursor.close()
    return desc, row_count


def sa_objs_to_dct(base, objs, relations=[]):
    """convert a queryset of objects into a list of dictionaries"""
    lstofdct = []
    for obj in objs:
        dct = sa_obj_to_dct(base, obj, relations=relations)
        try:
            dct["pk"] = obj.pk()
        except AttributeError:
            pass
        try:
            dct["pk_column"] = obj.pk_column().name
        except AttributeError:
            pass
        try:
            dct["ppk"] = obj.ppk()
        except AttributeError:
            pass
        try:
            dct["ppk_column"] = obj.ppk_column().name
        except AttributeError:
            pass
        lstofdct.append(dct)
    return lstofdct


def sa_obj_to_dct(base, obj, relations=[], depth=0):
    """
    Quick copy of NI's .row.  This will need to return all types in "primitives"
    Might need to integrate parts of serialize_obj
    """
    try:
        dct = obj.row()
    except AttributeError:
        dct = {}
        relations = copy.copy(relations)
        for key in obj._sa_class_manager.keys():
            #
            # for objects that are None assign the NullObject value in the row
            if key in relations:
                value = getattr(obj, key)
                relations.pop(relations.index(key))
                if isinstance(value, base):
                    # this is a single, call
                    dct[key] = sa_obj_to_dct(base, value, relations=relations, depth=depth + 1)
                elif isinstance(value, InstrumentedList):
                    # this is a _set
                    subvalue_lst = []
                    for subvalue in value:
                        subvalue_lst.append(sa_obj_to_dct(base, subvalue, relations=relations, depth=depth + 1))
                    dct[key] = subvalue_lst
                else:
                    # this somebody probably put the wrong relation in relations
                    # which would not work.
                    raise Exception("Unknown relation type %s" % type(value))
            else:
                value = getattr(obj, key)
                if isinstance(value, base):
                    dct[key] = value.pk()  # hack!
                elif isinstance(value, list):
                    # this would reverse lookup the entire database structure, lets
                    # not do that!
                    pass
                else:
                    dct[key] = value

            # dct[k] = value if (value or (k not in relations)) else obj.NoneObject()
    return dct
