# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

import copy
import functools
import logging
from pathlib import Path
import textwrap
from types import ModuleType
from typing import List, Optional
from collections.abc import Callable
import Octeres.embed.importer
import Octeres.embed.importer_restore_sys
from .source import get_all_imported_modules, format_comments, hash_source_text, indent_source_text, quote_source_text
from .types import ModuleSource, ModuleSourceSearchFunc, SourceText

log = logging.getLogger(__name__)


def find_embeddable_module_source(module_fullname: str, search_dirs: List[Path]) -> Optional[Path]:
    """
    Given an ordered set of directories in which to search, much like PYTHONPATH, search for a module's source code.

    FIXME: replace search_dirs with an ordered tuple of package directories and module files to be considered for inclusion
    """
    module_path = None
    for search_dir in search_dirs:
        search_path = Path(search_dir)
        module_layers = module_fullname.split(".")
        if search_path.name == module_layers[0] and Path(search_path, "__init__.py").is_file():
            search_path = search_path.parent
        module_base_path = Path(search_path, *module_layers)
        pkg_init_path = Path(module_base_path, "__init__.py")
        module_py_path = module_base_path.with_suffix(".py")
        if module_base_path.is_dir():
            if pkg_init_path.is_file():
                module_path = pkg_init_path
            elif module_py_path.is_file():
                module_path = module_py_path
            else:
                module_path = module_base_path
        elif module_py_path.is_file():
            module_path = module_py_path

        log.debug("Path for module %s: %s", module_fullname, str(module_path))

    return module_path


# def find_embeddable_modules(source: ModuleSource) -> ImportedModules:
#     """
#     Given an ordered set of directories in which to search, much like PYTHONPATH, find and return imported modules that are
#     embeddable.  Note: Python source code is required, which means that compiled code (.pyc) and extension modules will not be
#     considered embeddable.
#     """
#     module_paths: ImportedModulePaths = {}
#     find_all_module_source_code(imported_modules, search_dirs, module_paths)
#     embeddable_modules: ImportedModulePaths = {m: p for (m, p) in module_paths.items() if p is not None}

#     if log.isEnabledFor(logging.DEBUG):
#         msg_txt = textwrap.indent(pformat(embeddable_modules, sort_dicts=False), "    ")
#         log.debug("Embeddable modules:\n%s", msg_txt)

#     return embeddable_modules


def generate_embedded_module_code(source: ModuleSource) -> SourceText:
    # NOTE: this function has a side effect.  The text_hash attribute in source module data class is updated if not already set.
    is_pkg = source.path and source.path.name == "__init__.py"
    embedded_code: SourceText = [
        f"# Embedded code for module {source.module_fullname} from {str(source.path)}",
        f'_embedded_importer.add_module("{source.module_fullname}", {is_pkg}, "{str(source.path)}",',
    ]
    if not source.text_hash:
        source.text_hash = hash_source_text(source.text)
    text = copy.copy(source.text)
    text.extend(["", f"_embedded_importer_source_hash = '{source.text_hash}'"])
    text = quote_source_text(text)
    text[-1] += f', "{source.text_hash}")'
    embedded_code.extend(text)
    # TODO: optionally embed marshaled compiled code for the module to avoid compilation overhead.  Note: compiled code is
    # major/minor version specific.
    return embedded_code


_GENERATED_SCRIPT_HEADER = """
The code in this file is autogenerated and contains embedded modules.  Do not edit this file.  Instead, edit the original source
file and regenerate this embedded version.
"""


def _get_imported_module_source(module: ModuleType):
    get_source_fn: Callable[[str], str] = getattr(module.__loader__, "get_source")
    return get_source_fn(module.__name__).split("\n")


def generate_embedded_script(  # pylint: disable=unused-argument
    in_source: ModuleSource,
    search_dirs: List[Path],
    restore_sys: bool = False,
    new_context: bool = False,
    # TODO: add the ability to cache compiled code for a module to avoid repetitive compilation for embedded module code that has
    # not changed (based on hash).  time may also be a factor given long running processes may use multiple versions of the same
    # module, but different code, until they don't.
    cache_modules: bool = False,
    **kwargs,
) -> ModuleSource:
    in_text = in_source.text
    out_text: SourceText = []

    find_embeddable_module_source_fn: ModuleSourceSearchFunc = functools.partial(
        find_embeddable_module_source, search_dirs=search_dirs
    )
    imported_modules = get_all_imported_modules(in_source, find_embeddable_module_source_fn)

    if in_text[0].startswith("#!"):
        out_text.extend([in_text[0], ""])
        in_text.pop(0)
    while not in_text[0].strip():
        in_text.pop(0)
    out_text.extend(format_comments(_GENERATED_SCRIPT_HEADER))

    importer_text = _get_imported_module_source(Octeres.embed.importer)
    importer_cls_name = "_EmbeddedImporter"
    if restore_sys:
        importer_text.append("")
        importer_text.extend(_get_imported_module_source(Octeres.embed.importer_restore_sys))
        importer_cls_name = "_EmbeddedImporterRestoreSys"

    out_text.append("")
    if not new_context:
        out_text.extend(importer_text)
        out_text.append(f"_embedded_importer = {importer_cls_name}()")
    else:
        out_text.append("_embedded_importer_source = \\")
        out_text.extend(quote_source_text(importer_text))
        out_text.extend(
            [
                "",
                "_embedded_importer = eval(",
                f"    \"eval(compile(_embedded_importer_source, 'embedded_importer_source', 'exec')) or {importer_cls_name}()\",",
                '    {"_embedded_importer_source": _embedded_importer_source}, None)',
            ]
        )

    for imported_module in imported_modules.values():
        if imported_module.source is not None:
            out_text.append("")
            out_text.extend(generate_embedded_module_code(imported_module.source))
    out_text.append("")
    out_text.append("_embedded_importer.enable_importer()")

    out_text.extend(["", "try:"])
    out_text.extend(indent_source_text(in_text))
    out_text.extend(["finally:", "    _embedded_importer.cleanup_importer()"])

    if log.isEnabledFor(logging.DEBUG):
        msg_txt = textwrap.indent("\n".join(out_text), "    ").rstrip()
        log.debug("Generated script with embedded modules:\n%s", msg_txt)

    return ModuleSource(out_text)
