#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import pandas as pd
import numpy as np
import networkx as nx
from typing import Optional, List
from Octeres.bitmask.bitmask_graph_search import SearchStats, breadth_first_search


mask_lookup = {
    "&": np.logical_and.reduce,
    "|": np.logical_or.reduce,
    "!": np.logical_not.reduce,
    "^": np.logical_xor.reduce,
}
mask_lookup_extra = {
    "and": np.logical_and.reduce,
    "or": np.logical_or.reduce,
    "not": np.logical_not.reduce,
    "xor": np.logical_xor.reduce,
}
mask_lookup_extra.update(mask_lookup)
# v ^ the keys in one cannot be in the other for parsing
op_lookup = {
    "eq": np.equal,
    "==": np.equal,
    "neq": np.not_equal,
    "<>": np.not_equal,
    "gt": np.greater,
    ">": np.greater,
    "gte": np.greater_equal,
    ">=": np.greater_equal,
    "lt": np.less,
    "<": np.less,
    "lte": np.less_equal,
    "<=": np.less_equal,
    "in": np.isin,
    "not_in": lambda x, y: np.logical_and(np.isin(x, y)),
}
mop_lookup = dict(**op_lookup, **mask_lookup)


def parse_pair(pair):
    """split a pair on an operator."""
    real_mop = None
    for mop in mop_lookup.keys():
        if mop in pair:
            real_mop = mop
            break
    if not real_mop:
        raise Exception(f"mask operator not found in '{pair}'")
    a, b = pair.split(real_mop, 1)
    return {"name": a, "val": b, "op": real_mop}


def create_filt(logical_filt_str) -> dict:
    """create a filter from a string.
    Allowed to use the mask_lookup keys.
    """
    indexes = [(i if c == "(" else None, i if c == ")" else None) for i, c in enumerate(logical_filt_str) if c in ("(", ")")]
    if len(indexes) % 2 == 1:
        raise Exception("Unbalanced Parenthesis.")
    # the graph nodes names are index integers and -1 is the root.
    graph = nx.DiGraph()
    lookup = {}
    while indexes:
        new_indexes = []
        skip = False
        for i in range(total := len(indexes)):
            if skip:
                skip = False
                continue
            if (i + 1) == total:  # last
                a = indexes[i]  # leave it
                new_indexes.append(a)
            else:
                a = indexes[i]
                b = indexes[i + 1]
                al, ar = a
                bl, br = b
                if al is not None and ar is None and bl is None and br is not None:
                    p = None
                    if i - 1 < 0:
                        p = -1
                    else:
                        # search backwards for my parent.
                        # when searching backwards, increment for ) dec for (, then p
                        depth = 0
                        for j in range(i, 0, -1):
                            pl = indexes[j - 1][0]  # left
                            pr = indexes[j - 1][1]  # right
                            if pr is not None:  # found a right, need to find its pair
                                depth += 1
                            if pl is not None and depth == 0:  # found the correct level
                                p = pl
                                break
                            if pl is not None:  # found its pair
                                depth -= 1
                    if p is None:
                        raise Exception(f"Couldn't find parent for {al} {br}")
                    lookup[al] = (al, br, p)
                    graph.add_edge(p, al)  # node name is the start of the pair index
                    indexes[i] = (None, None)  # wipe them out so they are not parents
                    indexes[i + 1] = (None, None)
                    skip = True  # skip the next right bracket
                else:
                    new_indexes.append(a)
        if len(indexes) == len(new_indexes):
            raise Exception("Unmatchable parenthesis")
        indexes = new_indexes

    def search_func(parent_name: Optional[str], node_name: str, result: List[dict], debug=False) -> bool:
        return True

    def tail_func(node_name: str, results: List[dict], debug=False):
        rchar = "?"
        children = list(graph.successors(node_name))
        # children = nx.descendants(graph, node_name)
        if node_name == -1:
            # there should be only one child and it's 0
            assert len(children) == 1
            results.append(graph.nodes[0])
        elif children:
            i, j, p = lookup[node_name]
            cstr = logical_filt_str[:]
            for child in children:
                ci, cj, cp = lookup[child]
                cstr = cstr[:ci] + rchar * len(cstr[ci : cj + 1]) + cstr[cj + 1 :]

            # This is the mask operation.  Operation ordering is not implemented.
            op = cstr[i + 1 : j].replace(rchar, "")
            # this is very specific to mask_lookup and that each mask_lookup key len == 1
            op = "".join(set(op))
            dct = {}
            try:
                mask_lookup[op]
            except KeyError:
                raise Exception(f"invalid operation:{op}.  All mask operations within a parenthesis, must be the same.")
            else:
                # grab all the children's dictionaries and assemble the list with the operation.
                dct[op] = [graph.nodes[n] for n in children]
            graph.nodes[i].update(dct)
        else:
            i, j, p = lookup[node_name]
            dct = parse_pair(logical_filt_str[i + 1 : j])
            graph.nodes[i].update(dct)

    stat = SearchStats()
    results = breadth_first_search(-1, graph, stat, search_func, tail_func=tail_func, stop_when_found=False)
    assert len(results) == 1
    filt = results.pop()
    return filt


def create_mask(df: pd.DataFrame, logical_filt) -> np.ndarray:
    """given a filter as the result of create_filt, convert it to filter a pandas dataframe.
    Allowed to use the mask_lookup_extra keys.
    """
    if "name" in logical_filt:
        name = logical_filt["name"]
        val = logical_filt["val"]
        op = logical_filt["op"]
        if op in op_lookup:
            op_func = op_lookup[op]
            mask = op_func(df[name], val)
        else:
            raise Exception(f"Invalid Operation {op}")
    elif len(logical_filt) == 1:
        masks = []
        key = next(iter(logical_filt.keys()))
        value_lst = logical_filt[key]
        if key in mask_lookup_extra.keys():
            op_func = mask_lookup_extra[key]
            for sub_filt in value_lst:
                sub_mask = create_mask(df, sub_filt)
                masks.append(sub_mask)
            mask = op_func(masks)
        else:
            raise Exception(f"Invalid Operation {key}")
    else:
        raise Exception(f"Invalid Filter {logical_filt}")
    return mask
