#!/usr/bin/env python3
import gc
import time
from functools import wraps

import pandas as pd
import matplotlib.pyplot as plt
import memory_profiler
import random
import seaborn as sns
from python.Octeres.bitmask.bitmask_lite import (
    intervals_xor_set_v1,
    BitIntervals,
    intervals_or_v0,
    intervals_or_v4,
    intervals_or_v5,
    intervals_or_v6,
    intervals_xor_set_v0,
    intervals_xor_set_v2,
    intervals_xor_v0,
    intervals_xor_v5,
    intervals_and,
    intervals_and_v1,
    logical_xnor_v0,
    logical_xnor_v5,
    logical_xnor_v6,
    LiteBitmask,
    logical_xnor_v2,
    intervals_xor_v4,
    intervals_subtract,
    intervals_subtract_v1,
    intervals_subtract_v3,
    intervals_subtract_v4,
)

pd.options.mode.chained_assignment = "raise"
pd.set_option("display.max_rows", 1000)
pd.set_option("display.max_columns", 500)
pd.set_option("display.width", 1000)
pd.set_option("display.max_colwidth", 19)


def decorate_memory(func):
    @wraps(func)
    def decorate(*args, **kwargs):
        gc.disable()
        m_lst, result = memory_profiler.memory_usage((func, args, kwargs), interval=0.05, retval=True)
        gc.enable()
        assert result is not None
        return m_lst

    return decorate


def intervals_performance():
    profile_memory = False  # Parameter controlling whether to have memory test
    bit_mask = True  # Parameter controlling whether testing with bit_intervals or bitmask

    # A big test sample of intervals
    lst = [
        ([[i, i + 1] for i in range(0, 99999, 4)], [[i, i + 1] for i in range(2, 99999, 4)], "no_overlap_even"),
        ([[i, i + 1] for i in range(0, 99999)], [[i, i + 1] for i in range(0, 99999)], "full_overlap"),
        ([[i, i + 1] for i in range(0, 99999, 10)], [[i, i + 1] for i in range(5, 99999, 10)], "sparse_overlap"),
        (
            sorted([[i, i + 1] for i in random.sample(range(0, 99999), 50000)]),
            sorted([[i, i + 1] for i in random.sample(range(0, 99999), 50000)]),
            "random",
        ),
        ([[i, i + 1] for i in range(0, 99999, 2)], [[i, i + 1] for i in range(1, 99999, 2)], "alternating"),
        ([[i, i + 5] for i in range(0, 99999, 10)], [[i + 2, i + 3] for i in range(0, 99999, 10)], "nested"),
        ([[i, i + 1] for i in range(0, 99999, 3)], [[i, i + 1] for i in range(1, 99999, 4)], "mixed"),
        ([[i, i + 100] for i in range(0, 99999, 500)], [[i, i + 100] for i in range(250, 99999, 500)], "large_intervals"),
        ([[i, i + 1] for i in range(0, 99999, 3)], [[i, i + 1] for i in range(500, 99999, 3)], "3x3"),
        ([[i, i + 1] for i in range(0, 99999, 3)], [[i, i + 1] for i in range(500, 99999, 2)], "3x2"),
        ([[i, i + 1] for i in range(0, 99999)], [[i, i + 1] for i in range(1, 99999, 2)], "1x2"),
    ]
    """
    lst = [
        (
            [[i, i + 1] for i in range(0, 100000, 4)],
            [[i, i + 1] for i in range(2, 100000, 4)],
            "no_overlap_even",
        ),
        # Full overlap
        (
            [[i, i + 1] for i in range(0, 100000)],
            [[i, i + 1] for i in range(0, 100000)],
            "full_overlap",
        ),
    ]
    """

    def lots_of_memory(a, b):
        x = [0] * 1000000
        y = [1] * 1000000
        time.sleep(0.1)
        return True

    # A lot of test functions
    funcs = [
        # intervals_xor_set_v0,
        # intervals_xor_set_v1,
        # intervals_xor_set_v2,
        logical_xnor_v0,
        logical_xnor_v5,
        logical_xnor_v6,
        # logical_xnor_fast2,
        # intervals_xor_faster,
        # intervals_xor_copy,
        # intervals_xor_temp_interval,
        # intervals_xor_fastest3,
        # intervals_and,
        # intervals_or_v0,
        # intervals_or_v4,
        # intervals_or_v5,
        # intervals_or_v6,
        # intervals_or_v8,
        # intervals_and_fast,
        # intervals_subtract_fast4,
        # intervals_subtract_fast,
        # intervals_subtract_fast3
    ]
    # if profile_memory:
    #     funcs.append(lots_of_memory)

    count = 20
    results = []
    for alst, blst, speedcase in lst:
        for func in funcs:
            a_bit_intervals = BitIntervals.from_intervals_list(alst, validate=True)
            b_bit_intervals = BitIntervals.from_intervals_list(blst, validate=True)
            if bit_mask:
                # Convert the intervals to bitmasks
                a_bitmask = LiteBitmask(max(alst[-1][1] + 1, 99999 + 10), a_bit_intervals)
                b_bitmask = LiteBitmask(max(blst[-1][1] + 1, 99999 + 10), b_bit_intervals)
            warmup = True
            for i in range(count):
                ts = time.time()
                if bit_mask:
                    func(a_bitmask, b_bitmask)
                else:
                    func(a_bit_intervals, b_bit_intervals)
                te = time.time()
                gc.collect()
                if profile_memory:
                    mfunc = decorate_memory(func)
                    m = mfunc(a_bit_intervals, b_bit_intervals)
                    memory = max(m) - min(m)
                else:
                    memory = 0
                if warmup:
                    warmup = False
                    continue
                results.append(
                    {
                        "speedcase": speedcase,
                        "func": func.__name__,
                        "time(s)": te - ts,
                        "mem(Mb)": memory,
                        "i": i,
                    }
                )

    df = pd.DataFrame(results)
    dfg = df.groupby(["speedcase"])

    # Create a boxplot for the time
    plt.figure(figsize=(15, 8))
    sns.boxplot(x="speedcase", y="time(s)", hue="func", data=df)
    plt.xticks(rotation=45)
    plt.title("Execution Time by Function and Test Case")
    plt.tight_layout()
    plt.show()

    if profile_memory:
        # Create a boxplot for memory usage if profiling memory
        plt.figure(figsize=(15, 8))
        sns.boxplot(x="speedcase", y="mem(Mb)", hue="func", data=df)
        plt.xticks(rotation=45)
        plt.title("Memory Usage by Function and Test Case")
        plt.tight_layout()
        plt.show()

if __name__ == "__main__":
    intervals_performance()
    # bitmask_preformance()  # TODO
