#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import os
import sys
import argparse
import datetime
from dateutil.parser import parse as dp
import logging

parser = argparse.ArgumentParser(description="Search a log file for a section of time.")
parser.add_argument(
    "--range_format",
    dest="range_format",
    help=r"Range format Ex: 2018-03-05 15:26:21 -> %%Y-%%m-%%d %%H:%%M:%%S",
)
parser.add_argument("--range_start", dest="range_start", help="Start of the time cut.")
parser.add_argument("--range_end", dest="range_end", help="End of the time cut.")
parser.add_argument("file_names", metavar="F", type=str, nargs="+", help="File Name ")
parser.add_argument("--verbose", "-v", action="count")


def time_binary_search(log, file_name, range_format, range_start, range_end):
    """entry point for the recursive binary search"""
    file_stat = os.stat(file_name)
    file_size = file_stat.st_size

    # we need the length of a formatted line.
    now = datetime.datetime.now()
    now_str = now.strftime(range_format)
    format_length = len(now_str)
    log.debug("searching file %s" % file_name)
    log.debug("st_size:%s" % file_size)
    log.debug("%s Megabytes" % (file_size / (1024 * 1024)))

    with open(file_name, "r") as fd:
        pos_start = search(log, fd, format_length, range_start, 0, file_size)
        pos_end = search(log, fd, format_length, range_end, pos_start, file_size)

    log.debug("start:%s, end:%s" % (pos_start, pos_end))
    return pos_start, pos_end


def search(log, fd, format_length, ts, pos_start, pos_end, max_depth=32, depth=0, direction="-"):
    """binary search the file for the ts"""
    delta = pos_end - pos_start
    log.debug("%-16s %3s %16s    Δ:%16s" % (pos_start, direction, pos_end, delta))
    if delta == 1:
        return pos_start
    if depth > max_depth:
        # print("max_depth hit, returning start position")
        return pos_start
    pos_middle = pos_start + int((pos_end - pos_start) / 2)
    fd.seek(pos_middle)
    # this will most likely be a partial line
    line = fd.readline()
    # this should be the full line
    search_count = 128
    line_ts = None
    bad_lines = []
    while search_count > 0:
        line = fd.readline()
        line_ts_raw = line[:format_length]
        try:
            line_ts = dp(line_ts_raw)
        except ValueError:
            # keep searching
            search_count = search_count - 1
            bad_lines.append((fd.tell(), line.strip()))
        else:
            # we found it
            search_count = 0
    if line_ts is None:
        log.debug("*** We tried to find a date, but could not ***")
        for line_pos, line in bad_lines:
            log.debug("LP#%s:%s" % (line_pos, line))
        sys.exit(1)
    if line_ts < ts:
        # search forward
        pos = search(
            log,
            fd,
            format_length,
            ts,
            pos_middle,
            pos_end,
            max_depth=max_depth,
            depth=depth + 1,
            direction="->",
        )
    elif line_ts > ts:
        pos = search(
            log,
            fd,
            format_length,
            ts,
            pos_start,
            pos_middle,
            max_depth=max_depth,
            depth=depth + 1,
            direction="<-",
        )
    else:
        pos = pos_middle
    return pos


def read_cut(file_name, pos_start, pos_end):
    """read from start to end"""
    with open(file_name, "r") as fd:
        fd.seek(pos_start)
        line = fd.readline()
        pos = fd.tell()
        while line and pos < pos_end:
            print(line.strip())
            line = fd.readline()
            pos = fd.tell()


def main(args):
    logging.basicConfig()
    log = logging.getLogger("tcut")
    verbosity = args.verbose
    if verbosity is None:
        log.setLevel(logging.ERROR)
    if verbosity == 1:
        log.setLevel(logging.WARN)
    elif verbosity == 2:
        log.setLevel(logging.INFO)
    elif verbosity == 3:
        log.setLevel(logging.DEBUG)
    range_start = dp(args.range_start)
    range_end = dp(args.range_end)
    range_format = args.range_format
    file_names = args.file_names
    log.debug("Processing logs:\n")
    for file_name in file_names:
        log.debug("    %s" % file_name)
    log.debug("Cutting log from %s to %s" % (range_start, range_end))
    log.debug("Delta:%s hours" % ((range_end - range_start).total_seconds() / 3600.0))
    log.debug("Using format %s" % range_format)
    for file_name in file_names:
        pos_start, pos_end = time_binary_search(log, file_name, range_format, range_start, range_end)
        read_cut(file_name, pos_start, pos_end)
    log.debug("Cut complete.")


if __name__ == "__main__":
    args = parser.parse_args()
    main(args)
