#!/usr/bin/env python3

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
Only build-in and external libraries should be imported here.
"""
import datetime
import functools
import gc
import inspect
import os
import sys
import time
import traceback
from contextlib import contextmanager
from enum import Enum
from itertools import chain
from itertools import islice
from itertools import tee
from pprint import pformat
from types import FrameType
from typing import TypedDict, List, Union, Optional, Dict

from Octeres.forthwith import get_now, FORMAT_DATE_FULL, FORMAT_DATE_FULL_T


def memoize_fast(func):
    """memoization decorator, only memoizes args.  You must know what you
    are doing to use this function.
    """
    cache_dct = {}

    def function(*args):
        """the decorators function"""
        try:
            return cache_dct[args]
        except KeyError:
            results = func(*args)
            cache_dct[args] = results
            return results

    return function


def extract_traceback(include_time=True):
    """Extract a traceback, format it nicely and return it"""
    if include_time:
        currenttime = datetime.datetime.now().strftime(FORMAT_DATE_FULL)
    else:
        currenttime = ""
    (exc_cls, exc, tracbk) = sys.exc_info()
    exc_str = traceback.format_exception_only(exc_cls, exc)[0]
    tracebacklst = []
    tracebacklst.append(" ".join(("-" * 32, "START TRACEBACK", "-" * 30)))
    tracebacklst.append("    Exception  : %s" % (exc_str.strip()))
    tracebacklst.append("    Time       : %s" % currenttime)
    tracebacklst.append("-" * 80)
    stack = traceback.format_tb(tracbk)
    indent = "  "
    for stackpiece in stack:
        stackpiece = stackpiece.strip()
        stackpiece_lst = stackpiece.split(os.linesep)
        for error_line_idx, stack_item in enumerate(stackpiece_lst):
            tracebacklst.append(f"{indent}{currenttime}(L#{error_line_idx: >2})|{stack_item}")
    tracebacklst.append(" ".join(("-" * 32, "END TRACEBACK", "-" * 32)))
    return tracebacklst


def superformat(*msg, jump_back_count=0, parallel=False):
    """nearly the same as superprint, but does not print"""
    current_frame = inspect.currentframe()
    fname_lineno_lst = []
    local_count = jump_back_count + 1
    while local_count > 0:
        try:
            frame, filename, line_number, function_name, lines, index = inspect.getouterframes(current_frame)[1]
        except IndexError:
            break
        else:
            fpath, fname = os.path.split(filename)
            lineno = frame.f_lineno
            fname_lineno_lst.append((fname, lineno))
            current_frame = current_frame.f_back
            local_count -= 1
    now = get_now()
    now_str = now.strftime(FORMAT_DATE_FULL)
    if jump_back_count == 0:
        fname, lineno = fname_lineno_lst[0]
        if parallel:
            pairs = "%-13s|P#%s|L#%s" % (fname, os.getpid(), lineno)
        else:
            pairs = "%-13s:L#%s" % (fname, lineno)
    else:
        fname_lineno_lst.reverse()
        pairs = ""
        for fname, lineno in fname_lineno_lst:
            new_fname = fname if fname.find(".") == -1 else fname.split(".")[0]
            pairs += "%-16s<-" % ("%s.%s" % (new_fname, lineno))
        if parallel:
            pairs += "|P#%s|" % os.getpid()
    return "%s %s" % ("%s:%s>>>" % (pairs, now_str), " ".join(map(str, msg)))


def superprint(*msg, jump_back_count=0, parallel=False):
    """used to help find lost prints!"""
    current_frame = inspect.currentframe()
    fname_lineno_lst = []
    local_count = jump_back_count + 1
    while local_count > 0:
        try:
            frame, filename, line_number, function_name, lines, index = inspect.getouterframes(current_frame)[1]
        except IndexError:
            break
        else:
            fpath, fname = os.path.split(filename)
            lineno = frame.f_lineno
            fname_lineno_lst.append((fname, lineno))
            current_frame = current_frame.f_back
            local_count -= 1
    now = get_now()
    now_str = now.strftime(FORMAT_DATE_FULL)
    if jump_back_count == 0:
        fname, lineno = fname_lineno_lst[0]
        if parallel:
            pairs = "%-13s|P#%s|L#%s" % (fname, os.getpid(), lineno)
        else:
            pairs = "%-13s:L#%s" % (fname, lineno)
    else:
        fname_lineno_lst.reverse()
        pairs = ""
        for fname, lineno in fname_lineno_lst:
            new_fname = fname if fname.find(".") == -1 else fname.split(".")[0]
            pairs += "%-16s<-" % ("%s.%s" % (new_fname, lineno))
        if parallel:
            pairs += "|P#%s|" % os.getpid()
    print("%s:%s>>>" % (pairs, now_str), *msg)


def superinspect(**kwargs):
    for key, value in kwargs.items():
        superprint(
            f"{hex(id(value))}(R#{len(gc.get_referrers(value)): >4})(<>{str(type(value)): >16})(^{sys.getsizeof(value): >4}){key}={value}"
        )


def print_locals(filt="builtins", frame: FrameType = None, super=False):
    if frame:
        lcls = frame.f_back.f_locals
    else:
        lcls = inspect.currentframe().f_back.f_locals
    dct = {k: v for k, v in lcls.items() if v.__class__.__module__ == filt}
    s = ""
    for key, value in dct.items():
        if s:
            s += " "
        if type(value) == float:
            s += f"{key}={value:<8.3f}"
        elif type(value) == int:
            s += f"{key}={value: <8}"
        else:
            s += f"{key}={value}"
    if super:
        superprint(s)
    else:
        print(s)


def columnar_to_row(dct: Dict) -> List:
    """convert columnar to a list of dct"""
    lst = []
    lookup = {}
    for column, data_dct in dct.items():
        for idx, value in data_dct.items():
            try:
                lookup[idx]
            except KeyError:
                lookup[idx] = {}
            lookup[idx][column] = value
    index_lst = list(lookup.keys())
    index_lst.sort()
    for idx in index_lst:
        dct = lookup[idx]
        lst.append(dct)
    return lst


def df_to_lstofdct(df):
    """convert a dataframe to a list of dct"""
    df = df.reset_index()
    lst_of_dct = list(df.T.to_dict().values())
    return lst_of_dct


def auto_flush_stdout():
    """
    This forces realtime updating of stdout. No more flushing constantly for good information
    when running things on Jenkins.
    """
    sys.stdout = os.fdopen(sys.__stdout__.fileno(), "w", 1)
    global auto_flush_stdout

    def noop():
        pass

    auto_flush_stdout = noop


def iterate_chain(lst):
    """given an iterable, generate a new iterable that will
    provide the previous, current and next item."""
    prev_lst, main_lst, next_lst = tee(lst, 3)
    prev_lst = chain([None, None], prev_lst)
    main_lst = chain([None], main_lst, [None])
    next_lst = chain(next_lst, [None, None])
    return islice(zip(prev_lst, main_lst, next_lst), 1, len(lst) + 1)


class Timer:
    """a single timer to be used with Timed
    - Deprecated -
    """

    def __init__(self, timed: "Timed", name, use_gc=False):
        self.timed = timed
        self.name = name
        self.ts = -1.0
        self.te = -1.0
        self.td = -1.0
        self.use_gc = use_gc

    def __enter__(self):
        if self.use_gc is False:
            gc.disable()
        self.ts = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.te = time.time()
        self.td = self.te - self.ts
        self.timed.store(self)
        if self.use_gc is False:
            gc.enable()

    def result(self):
        return f"{self.name}:{self.td}"

    def get_dct(self):
        dct = dict(
            ts=self.ts,
            te=self.te,
            name=self.name,
            td=self.td,
        )
        return dct


class Timed:
    """class used to time a section of code using the with statement.

    Example:
    >>> from Octeres.util import Timed
    >>> import time
    >>> func = lambda x: time.sleep(x)
    >>> td = Timed()
    >>> with td.timer("sleep func"):
    ...     func(1)
    ...
    >>> td.results()
    ['sleep func:1.3589859008789062e-05', 'sleep func:1.002124309539795']
    >>>
    - Deprecated -
    """

    def __init__(self):
        self.times = []

    def timer(self, name, use_gc=True):
        """return a _Timer to be used with "with" """
        return Timer(self, name, use_gc=use_gc)

    def store(self, tm):
        self.times.append(tm)

    def results(self):
        return list(map(lambda tm: tm.result(), self.times))


def timed(print_func=None):
    """- Deprecated -"""

    def outer_decorator(func):
        """A decorator to time a function"""

        def inner_decorator(*args, **kwargs):
            """the decorators function"""
            start = time.time()
            ret = func(*args, **kwargs)
            end = time.time()
            if print_func is None:
                print("%s: %s" % (func.__name__, end - start))
            else:
                print_func("%s: %s" % (func.__name__, end - start))
            return ret

        return inner_decorator

    return outer_decorator


def truncate_message(data, size, location="left", message="*...*"):
    """given data, truncate the message to the size provided and
    add the truncated value inside that size."""
    locations = ["left", "right", "middle"]
    if location not in locations:
        raise Exception("invalid split location:%s" % location)
    len_truncated = len(message)
    if size < len_truncated * 2:
        raise Exception("Size is too small, please make it larger than %s" % (len_truncated * 2))

    len_data = len(data)
    if len_data <= size:
        result = data
    else:
        if location == "left":
            result = data[: size - len_truncated] + message
        elif location == "right":
            result = message + data[-(size - len_truncated) :]
        else:  # middle
            beginning = int((size - len_truncated) / 2)
            end = int((size - len_truncated) / 2 + ((size - len_truncated) % 2))
            result = data[:beginning] + message + data[-end:]

    return result


def time_pid_print(*args, **kwargs):
    print(
        datetime.datetime.now().strftime(FORMAT_DATE_FULL_T),
        os.getpid(),
        *args,
        **kwargs,
    )


def poll(total_count, max_loop_time):
    """
    Yield a loop sleep time for a count.
    max_loop_time = 0.25
    for sleep_time in poll(10, max_loop_time):
        ...
    """
    current_time = time.time()
    sleep_time = 0.0
    count = 0
    while count < total_count:
        if sleep_time > 0.0:
            time.sleep(sleep_time)
        yield sleep_time
        new_time = time.time()
        loop_time = new_time - current_time - sleep_time
        sleep_time = max_loop_time - loop_time
        count += 1
        current_time = new_time


def debug_with_rpdb():
    stdout = sys.stdout
    stderr = sys.stderr
    with open("/tmp/rpdb_stdout", "w") as fdout:
        sys.stdout = fdout
        with open("/tmp/rpdb_stderr", "w") as fderr:
            sys.stderr = fdout
            import rpdb

            rpdb.Rpdb().set_trace()
    sys.stdout = stdout
    sys.stderr = stderr
