# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import json
import operator
import pickle
from pprint import pformat
from typing import Callable, List
import pytest
from contextlib import nullcontext
from Octeres.bitmask import EnhancedBitmask
from Octeres.bitmask.bitmask_lite import (
    LiteBitmask,
    convert_bitmask_to_litebitmask,
    index_list_to_bitintervals,
    BitIntervals,
    intervals_xor,
)
from Octeres.bitmask.bitmask_lite import (
    convert_intervals_to_bitintervals,
    get_interval_bits,
    get_interval_bits_set,
    IntervalError,
)
from Octeres.bitmask.bitmask_lite import BitmaskFlags, intervals_or, intervals_and
from Octeres.serialization import DataEncoder, DataDecoder


fixtures_convert_intervals_to_bitintervals_00 = (
    pytest.param(
        [[1, 2], [4, 8]],
        nullcontext(enter_result=[[1, 2], [4, 8]]),
        nullcontext(enter_result=7),
    ),
    pytest.param([[1, 2], [1, 8]], pytest.raises(IntervalError), None),
    pytest.param([[1, 8], [1, 2]], pytest.raises(IntervalError), None),
    pytest.param([[1, 5], [3, 4]], pytest.raises(IntervalError), None),
    pytest.param([[1, 5], [3, 5]], pytest.raises(IntervalError), None),
    pytest.param([[1, 8]], nullcontext(enter_result=[[1, 8]]), nullcontext(enter_result=8)),
    pytest.param([[1, 8], [1, 8]], pytest.raises(IntervalError), None),
    pytest.param([[1, 2, 3]], pytest.raises(IntervalError), None),
    pytest.param([[]], pytest.raises(IntervalError), None),
    pytest.param([], nullcontext(enter_result=[]), nullcontext(enter_result=0)),
)


@pytest.mark.parametrize("intervals,exp_interval,exp_length", fixtures_convert_intervals_to_bitintervals_00)
def test_convert_intervals_to_bitintervals_00(intervals, exp_interval, exp_length):
    with exp_interval:
        bit_intervals = BitIntervals.from_intervals_list(intervals, validate=True)
        # bit_intervals = convert_intervals_to_bitintervals(intervals)
        assert bit_intervals.intervals == exp_interval.enter_result
        assert bit_intervals.bit_count == exp_length.enter_result
        assert len(list(get_interval_bits(bit_intervals))) == exp_length.enter_result
        assert len(list(get_interval_bits_set(bit_intervals))) == exp_length.enter_result


fixtures_LiteBitmask_init = (
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        17,
        (
            [
                [0, 2],
                [4, 7],
            ],
        ),
        nullcontext(enter_result=5),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        17,
        (
            [
                [0, 2, 1],
                [4, 7, 2],
            ],
        ),
        nullcontext(enter_result=4),
    ),
    pytest.param(LiteBitmask.zeros_and_set_bit, 17, (7,), nullcontext(enter_result=1)),
    pytest.param(LiteBitmask.zeros, 17, (), nullcontext(enter_result=0)),
    pytest.param(
        LiteBitmask.zeros_and_set_bits,
        17,
        ((0, 1, 9, 11, 8),),
        nullcontext(enter_result=5),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits,
        17,
        ((0, 1, 9, 11, 8, 0, 11),),
        nullcontext(enter_result=5),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits,
        19,
        ((1, 2, 8),),
        nullcontext(enter_result=3),
        id="bit tup",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits,
        19,
        ([1, 2, 8],),
        nullcontext(enter_result=3),
        id="bit lst",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        19,
        ([[1, 2], [4, 8]],),
        nullcontext(enter_result=5),
        id="bit ranges a1",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals,
        19,
        ([[1, 2], [3, 8]],),
        nullcontext(enter_result=8),
        id="bit ranges a2",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_intervals,
        19,
        ([[1, 2], [4, 8]],),
        nullcontext(enter_result=7),
        id="bit ranges a3",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        19,
        ([[4, 8, 2], [1, 8, 3]],),
        nullcontext(enter_result=4),
        id="bit ranges b",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        19,
        ([[1, 2], [1, 8, 3]],),
        nullcontext(enter_result=3),
        id="bit ranges c",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bit,
        19,
        (13,),
        nullcontext(enter_result=1),
        id="bit index",
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_ranges,
        4,
        ([[4, 8, 2], [1, 8, 3]],),
        pytest.raises(Exception),
        id="out of range",
    ),
)


@pytest.mark.parametrize("func,length,args,expectation", fixtures_LiteBitmask_init)
def test_LiteBitmask_init(func: Callable, length, args, expectation):
    with expectation:
        bitmask = func(length, *args)
        assert bitmask.bit_count == expectation.enter_result
        assert bitmask.length == length


def test_convert_bitmask_to_litebitmask():
    bit_lst = (1, 5, 7, 8)
    ebitmask: EnhancedBitmask = EnhancedBitmask.zeros_and_set_bits(17, bit_lst)
    lbitmask: LiteBitmask = convert_bitmask_to_litebitmask(ebitmask)
    e_bit_lst = ebitmask.get_idx_lst_from_bitmask(ebitmask)
    l_bit_lst = lbitmask.get_idx_lst_from_bitmask(lbitmask)
    assert e_bit_lst == l_bit_lst


def test_index_list_to_intervals_and_merge_00():
    a = [1, 4, 5, 6, 10, 15, 16, 17, 18, 22, 25, 26, 27, 28]
    b = [4, 6, 16, 17, 18, 22, 25, 27, 28]
    c = [[1, 1], [4, 6], [10, 10], [15, 18], [22, 22], [25, 28]]
    a_bit_intervals = index_list_to_bitintervals(a)
    b_bit_intervals = index_list_to_bitintervals(b)
    c_bit_intervals = BitIntervals.from_intervals_list(c)
    d_bit_intervals = intervals_or(a_bit_intervals, b_bit_intervals)
    assert c_bit_intervals == d_bit_intervals


def test_index_list_to_intervals_and_merge_01():
    a = [1, 4, 5, 6, 10, 15, 16, 17, 18, 22, 25, 26, 27, 28]
    b = [16, 17, 18, 22, 25, 27, 28]
    c = [[1, 1], [4, 6], [10, 10], [15, 18], [22, 22], [25, 28]]
    a_bit_intervals = index_list_to_bitintervals(a)
    b_bit_intervals = index_list_to_bitintervals(b)
    c_bit_intervals = BitIntervals.from_intervals_list(c)
    d_bit_intervals = intervals_or(a_bit_intervals, b_bit_intervals)
    assert c_bit_intervals == d_bit_intervals


# fixtures_intervals_count = (2**8)-1
fixtures_intervals_count = (2**19) - 1
fixtures_intervals = (
    pytest.param(
        intervals_or,
        [[5931, 5931]],
        [[5931, 5931]],
        nullcontext(enter_result=[[5931, 5931]]),
    ),
    pytest.param(intervals_or, [[1, 4]], [[2, 5]], nullcontext(enter_result=[[1, 5]])),
    pytest.param(intervals_or, [[1, 2]], [[4, 6]], nullcontext(enter_result=[[1, 2], [4, 6]])),
    pytest.param(intervals_or, [[4, 6]], [[1, 2]], nullcontext(enter_result=[[1, 2], [4, 6]])),
    pytest.param(
        intervals_or,
        [[4, 4], [5, 6]],
        [[1, 2]],
        nullcontext(enter_result=[[1, 2], [4, 6]]),
    ),
    pytest.param(intervals_or, [[0, 4]], [[2, 5]], nullcontext(enter_result=[[0, 5]])),
    pytest.param(intervals_or, [], [], nullcontext(enter_result=[])),
    pytest.param(intervals_or, [[0, 0]], [[0, 0]], nullcontext(enter_result=[[0, 0]])),
    pytest.param(
        intervals_or,
        [[i, i + 1] for i in range(999)],
        [[i, i + 1] for i in range(999)],
        nullcontext(enter_result=[[0, 999]]),
    ),
    pytest.param(
        intervals_or,
        [[i, i + 1] for i in range(0, 999, 3)],
        [[i, i + 1] for i in range(2, 999, 3)],
        nullcontext(enter_result=[[0, 999]]),
    ),
    pytest.param(
        intervals_or,
        [[i, i + 1] for i in range(0, 999, 4)],
        [[i, i + 1] for i in range(2, 999, 4)],
        nullcontext(enter_result=[[0, 999]]),
    ),
    pytest.param(
        intervals_and,
        [[5931, 5931]],
        [[5931, 5931]],
        nullcontext(enter_result=[[5931, 5931]]),
    ),
    pytest.param(intervals_and, [[1, 4]], [[2, 5]], nullcontext(enter_result=[[2, 4]])),
    pytest.param(intervals_and, [[1, 2]], [[4, 6]], nullcontext(enter_result=[])),
    pytest.param(intervals_and, [[4, 6]], [[1, 2]], nullcontext(enter_result=[])),
    pytest.param(intervals_and, [[4, 4], [5, 6]], [[1, 2]], nullcontext(enter_result=[])),
    pytest.param(intervals_and, [[0, 4]], [[2, 5]], nullcontext(enter_result=[[2, 4]])),
    pytest.param(intervals_and, [], [], nullcontext(enter_result=[])),
    pytest.param(intervals_and, [[0, 0]], [[0, 0]], nullcontext(enter_result=[[0, 0]])),
    pytest.param(
        intervals_and,
        [[i, i + 1] for i in range(999)],
        [[i, i + 1] for i in range(999)],
        nullcontext(enter_result=[[i, i + 1] for i in range(999)]),
    ),
    pytest.param(
        intervals_and,
        [[i, i + 1] for i in range(0, 999, 3)],
        [[i, i + 1] for i in range(2, 999, 3)],
        nullcontext(enter_result=[[i, i] for i in range(3, 999, 3)]),
    ),
    pytest.param(
        intervals_and,
        [[i, i + 1] for i in range(0, 999, 4)],
        [[i, i + 1] for i in range(2, 999, 4)],
        nullcontext(enter_result=[]),
    ),
    pytest.param(
        intervals_or,
        [[i, i + 1] for i in range(0, fixtures_intervals_count, 4)],
        [[i, i + 1] for i in range(2, fixtures_intervals_count, 4)],
        nullcontext(enter_result=[[0, fixtures_intervals_count]]),
    ),
    pytest.param(
        intervals_and,
        [[i, i + 1] for i in range(0, fixtures_intervals_count, 3)],
        [[i, i + 1] for i in range(2, fixtures_intervals_count, 3)],
        nullcontext(enter_result=[[i, i] for i in range(3, fixtures_intervals_count, 3)]),
    ),
)


@pytest.mark.parametrize("op,alst,blst,expectation", fixtures_intervals)
def test_intervals_00(op: Callable, alst: List[List], blst: List[List], expectation):
    with expectation:
        a_bit_intervals = BitIntervals.from_intervals_list(alst, validate=True)
        b_bit_intervals = BitIntervals.from_intervals_list(blst, validate=True)
        c_intervals = op(a_bit_intervals, b_bit_intervals)
        assert c_intervals.intervals == expectation.enter_result


def test_index_list_to_intervals_and_merge_02():
    # a = sorted([random.randint(0, 64) for _ in range(64)])
    a = [
        0,
        5,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        19,
        20,
        22,
        24,
        25,
        27,
        28,
        29,
        30,
        32,
        33,
        35,
        38,
        39,
        43,
        44,
        45,
        46,
        47,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        56,
        57,
        60,
        63,
    ]
    b = [
        2,
        3,
        4,
        5,
        7,
        8,
        10,
        11,
        13,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        26,
        32,
        33,
        34,
        35,
        37,
        39,
        40,
        41,
        43,
        45,
        46,
        49,
        50,
        54,
        55,
        57,
        59,
        60,
        61,
        62,
        63,
        64,
    ]
    c = BitIntervals.from_intervals_list([[0, 0], [2, 5], [7, 8], [10, 30], [32, 35], [37, 41], [43, 57], [59, 64]])
    a_bit_intervals = index_list_to_bitintervals(a)
    b_bit_intervals = index_list_to_bitintervals(b)
    assert a_bit_intervals.intervals == [
        [0, 0],
        [5, 5],
        [10, 17],
        [19, 20],
        [22, 22],
        [24, 25],
        [27, 30],
        [32, 33],
        [35, 35],
        [38, 39],
        [43, 54],
        [56, 57],
        [60, 60],
        [63, 63],
    ]
    assert b_bit_intervals.intervals == [
        [2, 5],
        [7, 8],
        [10, 11],
        [13, 13],
        [15, 24],
        [26, 26],
        [32, 35],
        [37, 37],
        [39, 41],
        [43, 43],
        [45, 46],
        [49, 50],
        [54, 55],
        [57, 57],
        [59, 64],
    ]
    r = intervals_or(a_bit_intervals, b_bit_intervals)
    assert c == r


fixtures_LiteBitmask = (
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [1, 2, 3]),
        # 01110000
        2,
        nullcontext(enter_result=True),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [0, 1, 2, 3, 7]),
        # 11110001
        slice(2, 8),
        nullcontext(enter_result=[True, True, False, False, False, True]),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [0, 1, 2, 3, 7]),
        # 11110001
        slice(2, 823042034),
        nullcontext(enter_result=[True, True, False, False, False, True]),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [1, 2, 3]),
        # 01110000
        slice(2, 5),
        nullcontext(enter_result=[True, True, False]),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [1, 2, 3]),
        # 01110000
        slice(4, 5),
        nullcontext(enter_result=[False]),
    ),
    pytest.param(
        LiteBitmask.zeros_and_set_bits(8, [1, 2, 3, 4]),
        # 01111000
        slice(0, 4, 2),
        nullcontext(enter_result=[False, True]),
    ),
)


@pytest.mark.parametrize("bitmask,slc,expectation", fixtures_LiteBitmask)
def test_LiteBitmask_00(bitmask, slc, expectation):
    with expectation:
        part_bitmask = bitmask[slc]
        assert expectation.enter_result == part_bitmask
        # tolist()


@pytest.mark.parametrize("method,length,args,expectation", fixtures_LiteBitmask_init)
def test_pickle(method: Callable, length, args, expectation):
    with expectation:
        bitmask = method(length, *args)
        assert bitmask.bit_count == expectation.enter_result
        pickle_str = pickle.dumps(bitmask)
        nbitmask = pickle.loads(pickle_str)
        assert nbitmask.bit_count == expectation.enter_result
        nbitmask.validate()


@pytest.mark.parametrize("method,length,args,expectation", fixtures_LiteBitmask_init)
def test_json(method: Callable, length, args, expectation):
    with expectation:
        bitmask: LiteBitmask = method(length, *args)
        assert bitmask.bit_count == expectation.enter_result
        json_str = json.dumps(bitmask, cls=DataEncoder)
        nbitmask = json.loads(json_str, cls=DataDecoder)
        assert nbitmask.bit_count == expectation.enter_result
        assert type(nbitmask.bit_intervals.intervals) == list
        bitmask.validate()
        nbitmask.validate()

        json_str2 = bitmask.to_json()
        bitmask2: LiteBitmask = LiteBitmask.from_json(json_str2)
        assert bitmask.bit_count == bitmask2.bit_count
        bitmask2.validate()


def test_json_BitmaskFlags():
    bf = BitmaskFlags()
    json_str = json.dumps(bf, cls=DataEncoder)
    bf2 = json.loads(json_str, cls=DataDecoder)
    assert bf.to_dict() == bf2.to_dict()

    bf = BitmaskFlags()
    bf.writeable = True
    json_str = json.dumps(bf, cls=DataEncoder)
    bf2 = json.loads(json_str, cls=DataDecoder)
    assert bf.to_dict() == bf2.to_dict()


fixtures_bitmasks_op = (
    pytest.param(
        operator.add,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.or_,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.add,
        8,
        (1, 2, 3),
        (1, 2, 3),
        nullcontext(enter_result=(1, 2, 3)),
        nullcontext(enter_result=(1, 2, 3)),
    ),
    pytest.param(
        operator.or_,
        8,
        (1, 2, 3),
        (1, 2, 3),
        nullcontext(enter_result=(1, 2, 3)),
        nullcontext(enter_result=(1, 2, 3)),
    ),
    pytest.param(
        operator.add,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.or_,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.eq,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(0, 2, 3, 5, 6, 7)),
        nullcontext(enter_result=(0, 2, 3, 5, 6, 7)),
    ),
    pytest.param(
        operator.eq,
        8,
        (1, 2, 3),
        (1, 2, 3),
        nullcontext(enter_result=(0, 1, 2, 3, 4, 5, 6, 7)),
        nullcontext(enter_result=(0, 1, 2, 3, 4, 5, 6, 7)),
    ),
    pytest.param(
        operator.eq,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(0, 5, 6, 7)),
        nullcontext(enter_result=(0, 5, 6, 7)),
    ),
    pytest.param(
        operator.sub,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1,)),
        nullcontext(enter_result=(4,)),
    ),
    pytest.param(
        operator.sub,
        8,
        (1, 2, 3),
        (1, 2, 3),
        nullcontext(enter_result=()),
        nullcontext(enter_result=()),
    ),
    pytest.param(
        operator.sub,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(2, 3, 4)),
        nullcontext(enter_result=(1,)),
    ),
    pytest.param(operator.sub, 8, (2, 3, 4), (9,), pytest.raises(Exception), None),
    pytest.param(operator.sub, 8, (2, 3, 4), (8,), pytest.raises(Exception), None),
    pytest.param(
        operator.sub,
        8,
        (2, 3, 4),
        (7,),
        nullcontext(enter_result=(2, 3, 4)),
        nullcontext(enter_result=(7,)),
    ),
    pytest.param(
        operator.or_,
        8,
        (2, 3, 4),
        (1,),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
    pytest.param(
        operator.or_,
        8,
        (2, 3, 4),
        (7,),
        nullcontext(enter_result=(2, 3, 4, 7)),
        nullcontext(enter_result=(2, 3, 4, 7)),
    ),
    pytest.param(
        LiteBitmask.logical_xor,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1, 4)),
        nullcontext(enter_result=(1, 4)),
    ),
    pytest.param(
        LiteBitmask.logical_xor,
        8,
        (0, 1, 2, 3),
        (0, 1, 2, 3),
        nullcontext(enter_result=()),
        nullcontext(enter_result=()),
    ),
    pytest.param(
        LiteBitmask.logical_xor,
        8,
        (0,),
        (3,),
        nullcontext(enter_result=(0, 3)),
        nullcontext(enter_result=(0, 3)),
    ),
    pytest.param(
        LiteBitmask.logical_xnor,
        8,
        (0,),
        (3,),
        nullcontext(enter_result=(1, 2, 4, 5, 6, 7)),
        nullcontext(enter_result=(1, 2, 4, 5, 6, 7)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 2, 6, 7),
        (1, 2, 3, 4, 5, 6, 7),
        nullcontext(enter_result=(1, 2, 6, 7)),
        nullcontext(enter_result=(1, 2, 6, 7)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(2, 3)),
        nullcontext(enter_result=(2, 3)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 2, 3),
        (1, 2, 3, 4),
        nullcontext(enter_result=(1, 2, 3)),
        nullcontext(enter_result=(1, 2, 3)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 4),
        (1, 4),
        nullcontext(enter_result=(1, 4)),
        nullcontext(enter_result=(1, 4)),
    ),
    pytest.param(
        LiteBitmask.logical_and,
        8,
        (1, 2),
        (6, 7),
        nullcontext(enter_result=()),
        nullcontext(enter_result=()),
    ),
    pytest.param(
        LiteBitmask.logical_or,
        8,
        (1, 2, 3),
        (2, 3, 4),
        nullcontext(enter_result=(1, 2, 3, 4)),
        nullcontext(enter_result=(1, 2, 3, 4)),
    ),
)


@pytest.mark.parametrize("op,length,abits,bbits,expectation_ab,expectation_ba", fixtures_bitmasks_op)
def test_LiteBitmasks_op(op: Callable, length: int, abits, bbits, expectation_ab, expectation_ba):
    with expectation_ab:
        abitmask = LiteBitmask.zeros_and_set_bits(length, abits)
        bbitmask = LiteBitmask.zeros_and_set_bits(length, bbits)
        assert abitmask.intersection_any(abitmask) is True
        assert abitmask.intersection_all(abitmask) is True
        abitmask.intersection_any(bbitmask)
        abitmask.intersection_all(bbitmask)
        cbitmask = op(abitmask, bbitmask)
        assert tuple(cbitmask.get_bits()) == expectation_ab.enter_result
        assert tuple(cbitmask.get_bits(bit_idx_first=0)) == expectation_ab.enter_result
        assert cbitmask.bit_count == len(expectation_ab.enter_result)
        assert cbitmask.length == length
        dbitmask = op(bbitmask, abitmask)
        assert tuple(dbitmask.get_bits()) == expectation_ba.enter_result
        assert dbitmask.bit_count == len(expectation_ba.enter_result)
        assert dbitmask.length == length
        for bitmask in [abitmask, bbitmask, cbitmask, dbitmask]:
            bitmask.validate()
            if bitmask.bit_count > 1:
                assert bitmask.intersection_any(bitmask) is True
        jbitmask = LiteBitmask.join_bitmasks([abitmask, bbitmask], op)
        assert tuple(jbitmask.get_bits()) == expectation_ab.enter_result
        assert jbitmask.bit_count == len(expectation_ab.enter_result)
        assert jbitmask.length == length
        hbitmask = LiteBitmask.join_bitmasks([bbitmask, abitmask], op)
        assert tuple(hbitmask.get_bits()) == expectation_ba.enter_result
        assert hbitmask.bit_count == len(expectation_ba.enter_result)
        assert hbitmask.length == length


fixtures_bitmasks_op_single = (pytest.param(LiteBitmask.logical_not, 8, (1, 2, 3), nullcontext(enter_result=(0, 4, 5, 6, 7))),)


@pytest.mark.parametrize("op,length,abits,expectation", fixtures_bitmasks_op_single)
def test_LiteBitmasks_op_single(op: Callable, length: int, abits, expectation):
    with expectation:
        abitmask = LiteBitmask.zeros_and_set_bits(length, abits)
        bbitmask = op(abitmask)
        assert tuple(bbitmask.get_bits()) == expectation.enter_result
        assert bbitmask.bit_count == len(expectation.enter_result)
        assert bbitmask.length == length


fixtures_bitmasks_op_truth = (
    pytest.param(
        LiteBitmask.intersection_all,
        8,
        (0,),
        (3,),
        nullcontext(enter_result=False),
        nullcontext(enter_result=False),
    ),
    pytest.param(
        LiteBitmask.intersection_all,
        8,
        (0,),
        (0,),
        nullcontext(enter_result=True),
        nullcontext(enter_result=True),
    ),
)


@pytest.mark.parametrize("op,length,abits,bbits,expectation_ab,expectation_ba", fixtures_bitmasks_op_truth)
def test_LiteBitmasks_op_truth(op: Callable, length: int, abits, bbits, expectation_ab, expectation_ba):
    with expectation_ab:
        abitmask = LiteBitmask.zeros_and_set_bits(length, abits)
        bbitmask = LiteBitmask.zeros_and_set_bits(length, bbits)
        atruth = op(abitmask, bbitmask)
        assert atruth == expectation_ab.enter_result
        btruth = op(bbitmask, abitmask)
        assert btruth == expectation_ba.enter_result
