# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import numpy as np
import pandas as pd
import pytest
from _pytest.assertion import truncate
from Octeres.forthwith import FORMAT_DATE_SECOND_T
from Octeres.logical import create_filt, create_mask


truncate.DEFAULT_MAX_LINES = 9999
truncate.DEFAULT_MAX_CHARS = 9999


def get_data():
    lstofdct = [
        {
            "machine_name": "mira",
            "te": "2013-04-09T00:00:00",
            "ts": "1970-01-01T00:00:00",
        },
        {
            "machine_name": "mira",
            "te": "2020-03-01T00:00:00",
            "ts": "2013-04-09T00:00:00",
        },
        {
            "machine_name": "mira",
            "te": "2038-01-01T00:00:00",
            "ts": "2020-03-01T00:00:00",
        },
        {
            "machine_name": "vesta",
            "te": "2013-04-09T00:00:00",
            "ts": "1970-01-01T00:00:00",
        },
        {
            "machine_name": "vesta",
            "te": "2013-04-10T00:00:00",
            "ts": "2013-04-09T00:00:00",
        },
        {
            "machine_name": "vesta",
            "te": "2038-01-01T00:00:00",
            "ts": "2013-04-10T00:00:00",
        },
        {
            "machine_name": "cetus",
            "te": "2013-04-09T00:00:00",
            "ts": "1970-01-01T00:00:00",
        },
        {
            "machine_name": "cetus",
            "te": "2013-04-10T00:00:00",
            "ts": "2013-04-09T00:00:00",
        },
        {
            "machine_name": "cetus",
            "te": "2038-01-01T00:00:00",
            "ts": "2013-04-10T00:00:00",
        },
    ]
    return lstofdct


def get_data_df():
    lstofdct = get_data()
    df = pd.DataFrame(lstofdct)
    df["ts"] = pd.to_datetime(df["ts"], format=FORMAT_DATE_SECOND_T)
    df["te"] = pd.to_datetime(df["te"], format=FORMAT_DATE_SECOND_T)
    return df


def get_filters(ids=False):
    filt_lst = [
        {
            "filt": "((machine_name==mira)&(te==2013-04-09))",
            "filt_result": {
                "&": [
                    {"name": "machine_name", "op": "==", "val": "mira"},
                    {"name": "te", "op": "==", "val": "2013-04-09"},
                ]
            },
            "mask_result": np.array([True, False, False, False, False, False, False, False, False]),
            "isvalid": True,
        },
        {
            "filt": "((machine_name==vesta)|((machine_name==mira)&(ts==2013-04-09)))",
            "filt_result": {
                "|": [
                    {"name": "machine_name", "op": "==", "val": "vesta"},
                    {
                        "&": [
                            {"name": "machine_name", "op": "==", "val": "mira"},
                            {"name": "ts", "op": "==", "val": "2013-04-09"},
                        ]
                    },
                ]
            },
            "mask_result": np.array([False, True, False, True, True, True, False, False, False]),
            "isvalid": True,
        },
        {
            "filt": "(((machine_name==theta)&(te==2013-04-09))|(((machine_name==mira)&(te==2013-04-10))|(te==2013-04-11)))",
            "filt_result": {
                "|": [
                    {
                        "&": [
                            {"name": "machine_name", "op": "==", "val": "theta"},
                            {"name": "te", "op": "==", "val": "2013-04-09"},
                        ]
                    },
                    {
                        "|": [
                            {"name": "te", "op": "==", "val": "2013-04-11"},
                            {
                                "&": [
                                    {"name": "machine_name", "op": "==", "val": "mira"},
                                    {"name": "te", "op": "==", "val": "2013-04-10"},
                                ]
                            },
                        ]
                    },
                ]
            },
            "mask_result": np.array([False, False, False, False, False, False, False, False, False]),
            "isvalid": True,
        },
        {
            "filt": "(((machine_name==theta)&(te==2013-04-09))|((machine_name==mira)&(te==2013-04-09)|(te==2013-04-09)))",
            "filt_result": {},
            "mask_result": None,
            "isvalid": False,
        },
        {
            "filt": "((machine_name==mira)and(te==2013-04-09))",
            "filt_result": {},
            "mask_result": None,
            "isvalid": False,
        },
    ]
    deck = (tuple(dct.values()) for dct in filt_lst)
    if ids:
        deck = [f"{i:0>2}" for i, _ in enumerate(deck)]
    return deck


@pytest.mark.parametrize("filt,filt_result,mask_result,isvalid", get_filters(), ids=get_filters(ids=True))
def test_create_filt(filt, filt_result, mask_result, isvalid):
    if isvalid:
        dct = create_filt(filt)
        assert dct == filt_result
    else:
        with pytest.raises(Exception) as err:
            create_filt(filt)


@pytest.mark.parametrize("filt,filt_result,mask_result,isvalid", get_filters(), ids=get_filters(ids=True))
def test_create_mask(filt, filt_result, mask_result, isvalid):
    df = get_data_df()
    if isvalid:
        dct = create_filt(filt)
        mask = create_mask(df, dct)
        np.testing.assert_equal(mask, mask_result)
        df = df[mask]


# # fixme: support no dataframe code.
# @pytest.mark.parametrize("filt,filt_result,mask_result,isvalid", get_filters(), ids=get_filters(ids=True))
# def test_create_lstmask(filt, filt_result, mask_result, isvalid):
#     lstofdct = get_data()
#     if isvalid:
#         dct = create_filt(filt)
#         assert 1 == 2
#         mask = create_lstmask(lstofdct, dct)
#         np.testing.assert_equal(mask, mask_result)
#         lstofdct = apply_mask(lstofdct, mask)
#         # FIXME: make sure the mask actually worked.
