#!/usr/bin/env python

# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

from pprint import pformat
import pytest
import pandas as pd
from pandas import to_datetime as pddt
from dateutil.parser import parse as dp

from Octeres.SL.modules.ltm import get_BitTree, get_Super_Location_to_Mask
from Octeres.bitmask import EnhancedBitmask, EnhancedBitmaskLike
from Octeres.bitmask.bitmask_lite import LiteBitmask
from Octeres.logical import create_filt
from Octeres.util import superprint, columnar_to_row
from Octeres.SL.abstract_location_to_mask import AbstractLocationtoMaskError

fixtures_bitmask_classes_dup = (
    pytest.param(EnhancedBitmask),
    pytest.param(LiteBitmask),
    pytest.param(EnhancedBitmask),  # hit it again to make sure it doesn't duplicate them.
    pytest.param(LiteBitmask),  # hit it again to make sure it doesn't duplicate them.
)


@pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes_dup)
def test_logical_filter_00(bitmask_class: EnhancedBitmaskLike):
    sltm = get_Super_Location_to_Mask(filt={"name": "test"}, bitmask_class=bitmask_class)
    logical_filt = {
        "&": [
            {"name": "name", "op": "==", "val": "test"},
            {"name": "alias", "op": "==", "val": "test_14_n"},
        ]
    }
    ltm_objs = sltm.ltm_object_factory_filter(logical_filt, include_test=True)
    assert sltm.bitmask_class == bitmask_class
    assert len(ltm_objs) == 1
    for ltm_obj in ltm_objs:
        assert type(ltm_obj.base_mask) == bitmask_class
        assert type(ltm_obj.graph.nodes["test"]["bitmask"]) == bitmask_class


def test_logical_filter_01():
    sltm = get_Super_Location_to_Mask(filt={"name": "test"})
    logical_filt = {
        "&": [
            {"name": "name", "op": "==", "val": "test"},
            {"name": "alias", "op": "==", "val": "TestEvents"},
        ]
    }
    ltm_objs = sltm.ltm_object_factory_filter(logical_filt, include_test=True)
    assert len(ltm_objs) == 5
    with pytest.raises(AbstractLocationtoMaskError):
        sltm.ltm_object_factory_filter_single(logical_filt, include_test=True)


def test_logical_filter_02():
    sltm = get_Super_Location_to_Mask(filt={"name": "test"})
    logical_filt = {
        "&": [
            {"name": "name", "op": "==", "val": "test"},
            {"name": "ts", "op": "==", "val": pddt("2011-07-01 00:00:00")},
            {"name": "te", "op": "==", "val": pddt("2011-07-01 00:02:00")},
        ]
    }
    ltm_objs = sltm.ltm_object_factory_filter(logical_filt, include_test=True)
    assert len(ltm_objs) == 1
    ltm_obj = sltm.ltm_object_factory_filter_single(logical_filt, include_test=True)
    assert ltm_obj is not None


def test_logical_filter_03():
    sltm = get_Super_Location_to_Mask(filt={"name": "test"})
    logical_filt = {
        "&": [
            {"name": "name", "op": "==", "val": "test"},
            {"name": "alias", "op": "==", "val": "testflat8"},
        ]
    }
    ltm_objs = sltm.ltm_object_factory_filter(logical_filt, include_test=True)
    assert len(ltm_objs) == 1
    ltm_obj = sltm.ltm_object_factory_filter_single(logical_filt, include_test=True)
    assert ltm_obj is not None


@pytest.mark.parametrize("bitmask_class", fixtures_bitmask_classes_dup)
def test_logical_filter_04(bitmask_class: EnhancedBitmaskLike):
    sltm = get_Super_Location_to_Mask(filt={"name": "test"}, bitmask_class=bitmask_class)
    logical_filt = {
        "&": [
            {"name": "name", "op": "==", "val": "test"},
        ]
    }
    ltm_objs = sltm.ltm_object_factory_filter(logical_filt, include_test=True)
    for ltm_obj in ltm_objs:
        assert type(ltm_obj.base_mask) == bitmask_class
        assert type(ltm_obj.graph.nodes["test"]["bitmask"]) == bitmask_class


def test_sltm_00():
    sltm = get_Super_Location_to_Mask(filt={"name": "test"})
    len_df = len(sltm._df)


def test_bittree_00():
    logical_filt_str = "(((name==test)&(te==2012-02-04))|((name==test)&((te==2012-02-14)|(te==2012-02-16))))"
    logical_filt = create_filt(logical_filt_str)
    print(pformat(logical_filt))
    df = get_BitTree(logical_filt)
    print(df)
    df = pd.DataFrame(df, columns=["machine_name", "ts", "te"])
    df["ts"] = df["ts"].apply(lambda t: t.isoformat())
    df["te"] = df["te"].apply(lambda t: t.isoformat())
    lstofdct = columnar_to_row(df)
    print(pformat(lstofdct))
    assert df.empty is False
