# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import re
import datetime
import pandas as pd
import networkx as nx
from Octeres.SL.abstract_location_to_mask import (
    add_ghost_node,
    get_bitmask_from_locations,
    get_ltm_from_graph,
)
from Octeres.SL.modules.ltm import get_Super_Location_to_Mask
from Octeres.util import superprint


class TestLTM:
    machine_name = "salaf"

    def test_00(self):
        ts = datetime.datetime(2021, 8, 25)
        sltm = get_Super_Location_to_Mask(filt={"name": self.machine_name, "pit": ts})
        df = sltm.get_ltm_df({"name": self.machine_name, "pit": pd.to_datetime(ts)})
        assert df.empty is False
        ltm_obj = sltm.ltm_object_factory(self.machine_name, ts)

        assert (match := re.match(ltm_obj.re_row, "x30")) is not None, match
        assert (match := re.match(ltm_obj.re_row, "x3")) is None, match
        assert (match := re.match(ltm_obj.re_row, "x300")) is None, match
        assert (match := re.match(ltm_obj.re_row, "x")) is None, match

        assert (match := re.match(ltm_obj.re_rack, "x3000")) is not None, match
        assert (match := re.match(ltm_obj.re_rack, "x30000")) is None, match
        assert (match := re.match(ltm_obj.re_rack, "x")) is None, match
        assert (match := re.match(ltm_obj.re_rack, "x3")) is None, match

        assert (match := re.match(ltm_obj.re_chassis, "x3000c0")) is not None, match
        assert (match := re.match(ltm_obj.re_chassis, "x30000c10")) is None, match
        assert (match := re.match(ltm_obj.re_chassis, "x30000c")) is None, match
        assert (match := re.match(ltm_obj.re_chassis, "x30000")) is None, match

        assert (match := re.match(ltm_obj.re_slot, "x3000c0s0")) is not None, match
        assert (match := re.match(ltm_obj.re_slot, "x3000c0s10")) is None, match
        assert (match := re.match(ltm_obj.re_slot, "x3000c0s")) is None, match
        assert (match := re.match(ltm_obj.re_slot, "x3000c0")) is None, match

        assert (match := re.match(ltm_obj.re_blade, "x3000c0s0b0")) is not None, match
        assert (match := re.match(ltm_obj.re_blade, "x3000c0s0b10")) is None, match
        assert (match := re.match(ltm_obj.re_blade, "x3000c0s0b")) is None, match
        assert (match := re.match(ltm_obj.re_blade, "x3000c0s0")) is None, match

        assert (match := re.match(ltm_obj.re_node, "x3000c0s0b0n1")) is not None, match
        assert (match := re.match(ltm_obj.re_node, "x3000c0s0b0n10")) is None, match
        assert (match := re.match(ltm_obj.re_node, "x3000c0s0b0n")) is None, match
        assert (match := re.match(ltm_obj.re_node, "x3000c0s0b0")) is None, match

        assert (parents := ltm_obj.get_parents(self.machine_name, "x3000c0s0b0n0")) == ["x3000c0s0b0"], parents
        assert (parents := ltm_obj.get_parents(self.machine_name, "x3000c0s0b0")) == ["x3000c0s0"], parents
        assert (parents := ltm_obj.get_parents(self.machine_name, "x3000c0s0")) == ["x3000c0"], parents
        assert (parents := ltm_obj.get_parents(self.machine_name, "x3000c0")) == ["x3000"], parents
        assert (parents := ltm_obj.get_parents(self.machine_name, "x3000")) == ["x30"], parents
        assert (parents := ltm_obj.get_parents(self.machine_name, "x30")) == [self.machine_name], parents

        # assert (match := re.match(LTM.re_cgroup, 'x3000c0s0b0n1[0]')) is not None, match
        # assert (match := re.match(LTM.re_cgroup, 'x3000c0s0b0n1[10]')) is None, match
        # assert (match := re.match(LTM.re_cgroup, 'x3000c0s0b0n1[]')) is None, match
        # assert (match := re.match(LTM.re_cgroup, 'x3000c0s0b0n1')) is None, match

        # assert (parents := LTM.get_parents(self.machine_name, 'x3000c0s0b0n0[3]')) == ['x3000c0s0b0n0'], parents

    def test_01(self):
        """this is going the round about way to get the ltm_dct.  This can be generated from get_ltm_from_graph using ltm_class
        as the bitmask is already there."""
        range_start = datetime.datetime(2021, 8, 26)
        sltm = get_Super_Location_to_Mask(filt={"name": self.machine_name, "pit": range_start})
        ltm_obj = sltm.ltm_object_factory(self.machine_name, range_start)
        range_end = datetime.datetime(2021, 8, 27)
        superprint("get_ltm_from_graph")
        ltm_dct = get_ltm_from_graph(
            ltm_obj.graph,
            ltm_obj.bit_total,
            self.machine_name,
            range_start,
            range_end,
            ltm_obj.get_parents,
            ltm_obj.get_obj_type,
            ltm_obj.get_is_schedulable,
            debug=False,
        )
        # show_nxgraph(graph)

    def test_02(self):
        ts = datetime.datetime(2021, 8, 25)
        sltm = get_Super_Location_to_Mask(filt={"name": self.machine_name, "pit": ts})
        ltm_obj = sltm.ltm_object_factory(self.machine_name, ts)
        graph = nx.DiGraph()
        add_ghost_node(
            graph,
            self.machine_name,
            self.machine_name,
            ltm_obj.get_parents,
            ltm_obj.get_obj_type,
        )
        add_ghost_node(
            graph,
            self.machine_name,
            "x3000c0s0b0n0",
            ltm_obj.get_parents,
            ltm_obj.get_obj_type,
        )
        add_ghost_node(
            graph,
            self.machine_name,
            "x3001c6s0b1n0",
            ltm_obj.get_parents,
            ltm_obj.get_obj_type,
        )

    def test_03(self):
        point_in_time = datetime.datetime(2021, 8, 25)
        sltm = get_Super_Location_to_Mask(filt={"name": self.machine_name})
        ltm_obj = sltm.ltm_object_factory(self.machine_name, point_in_time)
        superprint(get_bitmask_from_locations(ltm_obj, "node_name_short", ["x3203c5s0b0n0"]))
        superprint(get_bitmask_from_locations(ltm_obj, "", ["x3203c5s0b0n0"]))
